"""
Utility functions
"""
from __future__ import division
from builtins import range
import functools
import warnings


def average_length(a1s, a1e, a2s, a2e):
    """
    Given two sets of coordinates, a1 and a2, returns the average length.

    :param int a1s: a1 leftmost number
    :param int a1e: a1 rightmost number
    :param int a2s: a2 leftmost number
    :param int a2e: a2 rightmost number

    :return float: the average length
    """
    return ((a1e - a1s) + (a2e - a2s) + 2) / 2.0


def between(pos, start, end):
    """
    Tests if a number is between two others

    :param int pos: number to test
    :param int start: leftmost number
    :param int end: rightmost number

    :return bool: if the number is between start and end
    """
    if pos < start or pos > end:
        return False
    return True


def union_range(start1, end1, start2, end2):
    """
    .. versionadded:: 0.1.12

    .. versionchanged:: 0.3.1
        changed behaviour, since the intervals are meant to be closed

    If two numeric ranges overlap, it returns the new range, otherwise None is
    returned. Works on both int and float numbers, even mixed.

    Arguments:
        start1 (numeric): start of range 1
        end1 (numeric): end of range 1
        start2 (numeric): start of range 2
        end2 (numeric): end of range 2

    Returns:
        (tuple or None): union of the ranges or None if the ranges don't
        overlap

    Example:
        >>> union_range(10, 13, 1, 10)
        (1, 13)
        >>> union_range(1, 10, 11, 13)
        (1, 13)
        >>> union_range(1, 10, 12, 13)
        None

    """
    if between(start2, start1, end1) or \
        between(end2, start1, end1) or \
            (end1 + 1 == start2) or (end2 + 1 == start1):
        return min(start1, start2), max(end1, end2)
    return None


def union_ranges(intervals):
    """
    .. versionadded:: 0.3.1

    From a list of ranges, assumed to be closed, performs a union of all
    elements.

    Arguments:
        intervals (intervals): iterable where each element is a closed range
            (tuple)

    Returns:
        list: the list of ranges that are the union of all elements passed

    Examples:
        >>> union_ranges([(1, 2), (3, 7), (6, 12), (9, 17), (18, 20)])
        [(1, 20)]
        >>> union_ranges([(1, 2), (3, 7), (6, 12), (9, 14), (18, 20)])
        [(1, 14), (18, 20)]
    """
    intervals = sorted(intervals)

    union = [intervals.pop(0)]

    for start2, end2 in intervals:
        start1, end1 = union[-1]
        new_range = union_range(start1, end1, start2, end2)
        if new_range is None:
            union.append(
                (start2, end2)
            )
        else:
            union[-1] = new_range
    return union


def complement_ranges(intervals, end=None):
    """
    .. versionadded:: 0.3.1

    Perform a complement operation of the list of intervals, i.e. returning the
    ranges (tuples) that are not included in the list of intervals.
    :func:`union_ranges` is first called on the intervals.

    .. note::

        the `end` parameter is there for cases where the ranges passed don't
        cover the whole space. Assuming a list of ranges from annotations on a
        nucleotidic sequence, if the last range doesn't include the last
        position of the sequence, passing end equal to the length of the
        sequence will make the function include a last range that includes it

    Arguments:
        intervals (intervals): iterable where each element is a closed range
            (tuple)
        end (int): if the end of the complement intervals is supposed to be
            outside the last range.

    Returns:
        list: the list of intervals that complement the ones passed.

    Examples:
        >>> complement_ranges([(1, 10), (11, 20), (25, 30)], end=100)
        [(21, 24), (31, 100)]
        >>> complement_ranges([(1, 10), (11, 20), (25, 30)])
        [(21, 24)]
        >>> complement_ranges([(0, 2), (3, 17), (18, 20)])
        []
        >>> complement_ranges([(0, 2), (3, 17), (18, 20)], end=100)
        [(21, 100)]
    """
    intervals = union_ranges(intervals)

    comp_intervals = []

    if intervals[0][0] > 1:
        comp_intervals.append((1, intervals[0][0] - 1))

    for index in range(0, len(intervals) - 1):
        new_start = intervals[index][1] + 1
        new_end = intervals[index + 1][0] - 1
        if new_start < new_end:
            comp_intervals.append(
                (new_start, new_end)
            )

    if end is not None:
        start = intervals[-1][1] + 1
        if end > start:
            comp_intervals.append((start, end))

    return comp_intervals


def ranges_length(ranges):
    """
    .. versionadded:: 0.1.12

    Given an iterable where each element is a range, a tuple whose elements
    are numbers with the first being less than or equal to the second, the
    function sums the lengths of all ranges.

    .. warning::

        it's supposed to be used on intervals that were first passed to
        functions like :func:`union_ranges`. If values overlap, there the sum
        will be wrong

    Arguments:
        ranges (iterable): each element is a tuple like *(1, 10)*

    Returns:
        int: sum of all ranges lengths

    """
    return sum(range[1] - range[0] + 1 for range in ranges)


def range_substract(start1, end1, start2, end2):
    intersect = range_intersect(start1, end1, start2, end2)

    if intersect is None:
        return [(start1, end1)]

    ranges = []
    if start1 != intersect[0]:
        ranges.append((start1, intersect[0] - 1))
    if end1 != intersect[1]:
        ranges.append((intersect[1] + 1, end1))
    return ranges


def range_intersect(start1, end1, start2, end2):
    """
    .. versionadded:: 0.1.13

    Given two ranges in the form *(start, end)*, it returns the range
    that is the intersection of the two.

    Arguments:
        start1 (int): start position for the first range
        end1 (int): end position for the first range
        start2 (int): start position for the second range
        end2 (int): end position for the second range

    Returns:
        (None, tuple): returns a tuple with the start and end position for
        the intersection of the two ranges, or *None* if the intersection is
        empty
    """
    if between(start2, start1, end1) or between(end2, start1, end1) or \
       between(start1, start2, end2) or between(end1, start2, end2):
        return max(start1, start2), min(end1, end2)
    return None


def apply_func_window(func, data, window, step=0):

    if step == 0:
        step = window

    for index in range(0, len(data), step):
        yield func(data[index:index+window])


def deprecated(func):
    '''
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.

    from https://wiki.python.org/moin/PythonDecoratorLibrary
    '''

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn_explicit(
            "Call to deprecated function {}.\n{}".format(
                func.__name__,
                func.__doc__
            ),
            category=DeprecationWarning,
            filename=func.__code__.co_filename,
            lineno=func.__code__.co_firstlineno + 1
        )
        return func(*args, **kwargs)
    return new_func



"""
.. versionadded:: 0.1.14

EC mappings
"""
from future.utils import viewitems
import re
from ..io import open_file

LEVEL1_NAMES = {
    1: 'oxidoreductases',
    2: 'transferases',
    3: 'hydrolases',
    4: 'lyases',
    5: 'isomerases',
    6: 'ligases'
}
"""
Top level classification names
"""

ENZCLASS_REGEX = r"^(\d)\. ?([\d-]+)\. ?([\d-]+)\. ?([\d-]+) +(.+)\."
"""
Used to get the description for the higher level enzyme classes from the file
*enzclass.txt* on `expasy <http://expasy.org>`_
"""


def parse_expasy_file(file_name):
    """
    .. versionchanged:: 0.4.2
        changed to work on python 3.x

    Used to load enzyme descriptions from the file *enzclass.txt* on
    `expasy <http://expasy.org>`_.

    The FTP url for enzclass.txt is:
    `<ftp://ftp.expasy.org/databases/enzyme/enzclass.txt>`_
    """
    labels = {}

    for line in open_file(file_name, mode='rb'):
        line = line.decode('ascii')
        match = re.search(ENZCLASS_REGEX, line)

        if match is None:
            continue

        ec = '.'.join(value for value in match.groups()[:4] if value != '-')

        labels[ec] = match.group(5)

    return labels


def get_enzyme_level(ec, level=4):
    """
    .. versionadded:: 0.1.14

    Returns an enzyme class at a specific level , between 1 and 4 (by default
    the most specific, 4)

    Arguments:
        ec (str): a string representing an EC number (e.g. 1.2.4.10)
        level (int): from 1 to 4, to get a different level specificity of in
            the enzyme classification

    Returns:
        str: the EC number at the requested specificity

    Example:
        >>> from mgkit.mappings.enzyme import get_enzyme_level
        >>> get_enzyme_level('1.1.3.4', 1)
        '1'
        >>> get_enzyme_level('1.1.3.4', 2)
        '1.1'
        >>> get_enzyme_level('1.1.3.4', 3)
        '1.1.3'
        >>> get_enzyme_level('1.1.3.4', 4)
        '1.1.3.4'
    """
    return '.'.join(ec.split('.')[:level])


def change_mapping_level(ec_map, level=3):
    """
    .. versionadded:: 0.1.14

    Given a dictionary, whose values are dictionaries, in which a key is named
    *ec* and its value is an iterable of EC numbers, returns an iterator that
    can be used to build a dictionary with the same top level keys and the
    values are sets of the transformed EC numbers.

    Arguments:
        ec_map (dict): dictionary generated by
            :func:`mgkit.net.uniprot.get_gene_info`
        level (int): number from 1 to 4, to specify the level of the mapping,
            passed to :func:`get_enzyme_level`

    Yields:
        tuple: a tuple (gene_id, set(ECs)), which can be passed to *dict* to
        make a dictionary

    Example:
        >>> from mgkit.net.uniprot import get_gene_info
        >>> from mgkit.mappings.enzyme import change_mapping_level
        >>> ec_map = get_gene_info('Q9HFQ1', columns='ec')
        {'Q9HFQ1': {'ec': '1.1.3.4'}}
        >>> dict(change_mapping_level(ec_map, level=2))
        {'Q9HFQ1': {'1.1'}}

    """
    for gene_id, ecdict in viewitems(ec_map):
        try:
            ec_list = ecdict['ec']
        except KeyError:
            continue

        if isinstance(ec_list, str):
            ec_list = [ec_list]

        yield gene_id, set(
            get_enzyme_level(ec_id, level=level) for ec_id in ec_list
        )


def get_mapping_level(ec_map, level=3):
    """
    .. versionadded:: 0.3.0

    Given a dictionary, whose values are iterable of EC numbers, returns an
    iterator that can be used to build a dictionary with the same top level
    keys and the values are sets of the transformed EC numbers.

    Arguments:
        ec_map (dict): dictionary genes to EC
        level (int): number from 1 to 4, to specify the level of the mapping,
            passed to :func:`get_enzyme_level`

    Yields:
        tuple: a tuple (gene_id, set(ECs)), which can be passed to *dict* to
        make a dictionary
    """
    for gene_id, ec_list in viewitems(ec_map):

        if not ec_list:
            continue

        if isinstance(ec_list, str):
            ec_list = [ec_list]

        yield gene_id, set(
            get_enzyme_level(ec_id, level=level) for ec_id in ec_list
        )


def get_enzyme_full_name(ec_id, ec_names, sep=', '):
    """
    .. versionadded:: 0.2.1

    From a EC identifiers and a dictionary of names builds a comma separated
    name (by default) that identifies the function of the enzyme.

    Arguments:
        ec_id (str): EC identifier
        ec_names (dict): a dictionary of names that can be produced using
            :func:`parse_expasy_file`
        sep (str): string used to join the names

    Returns:
        str: the enzyme classification name
    """

    name_list = []

    while True:
        try:
            name_list.append(
                ec_names[ec_id]
            )
        except KeyError:
            pass

        ec_id = '.'.join(ec_id.split('.')[:-1])

        if not ec_id:
            break

    return sep.join(reversed(name_list))


def parse_expasy_dat_section(expasy_dat_section, skip_comments=True, skip_codes=None):
    """
    .. versionadded:: 0.4.2

    Parses an entry of the `enzyme.dat` file in expasy, used internally by
    :func:`mgkit.mappings.enzyme.parse_expasy_dat`, with the other arguments
    being passed over from it.

    Returns:
        dict: dictionary with the entry, with keys being the codes of the entry
        and the values the lines
    """
    parsed_lines = {}
    for line in expasy_dat_section:
        if skip_comments and line.startswith('CC'):
            continue
        key, value = line.rstrip().split(maxsplit=1)
        if (skip_codes is not None) and (key in skip_codes):
            continue
        try:
            parsed_lines[key].append(value)
        except KeyError:
            parsed_lines[key] = [value]

    if 'CA' in parsed_lines:
        ca_lines = ''.join(parsed_lines['CA'])
        parsed_lines['CA'] = ca_lines.rstrip('.').split('. ')

    return parsed_lines

def parse_expasy_dat(expasy_dat, keep_empty=False, skip_comments=True, skip_codes=None):
    """
    .. versionadded:: 0.4.2

    Parses the information in `enzyme.dat` file in expasy, a flat file
    containting the information about the enzyme classification.

    It can be downloaded at:
    `<ftp://ftp.expasy.org/databases/enzyme/enzyme.dat>`_

    Arguments:
        expasy_dat (str): file name or handle to an expasy.dat file
        keep_empty (bool): section that are empty are removed by default
        skip_comments (bool): used to avoid returning comments (lines starting)
            with `CC` in the file
        skip_codes (set, tuple): set or tuple or list to skip specific parts
            of the file, like `skip_comments`
    Yields:
        dict: dictionary with each entry in the file, where the keys are the
        codes and the values are the lines included in the file
    """
    section = []
    for line in open_file(expasy_dat, 'rb'):
        line = line.decode('ascii')

        # section end
        if line.startswith('//'):
            parsed_section = parse_expasy_dat_section(section, skip_comments=skip_comments, skip_codes=skip_codes)
            section = []
            if parsed_section or keep_empty:
                yield parsed_section
            continue

        section.append(line)



"""
.. versionadded:: 0.1.14

Misc code

"""
from builtins import zip
import logging

from matplotlib.gridspec import GridSpec
from matplotlib import patches as mpatches

LOG = logging.getLogger(__name__)


def get_single_figure(dpi=300, figsize=(10, 20), aspect='auto'):
    """
    .. versionchanged:: 0.1.14
        added *aspect* parameter

    Simple wrapper to init a single figure

    Arguments:
        dpi (int): dpi used for the figure
        figsize (tuple): size of the figure in inches
        aspect (str, float): aspect ratio to be passed to figure.add_subplot

    Returns:
        tuple: the figure and axes objects
    """
    import matplotlib.pyplot
    fig = matplotlib.pyplot.figure(dpi=dpi, figsize=figsize)
    ax = fig.add_subplot(111, aspect=aspect)
    return fig, ax


def get_grid_figure(rows, cols, dpi=300, figsize=(10, 20), **kwd):
    """
    .. versionadded:: 0.1.13

    Simple wrapper to init a GridSpec figure

    Arguments:
        rows (int): number of rows
        columns (int): number of columns
        dpi (int): dpi used for the figure
        figsize (tuple): size of the figure in inches

    Returns:
        tuple: the figure and axes objects
    """
    import matplotlib.pyplot
    fig = matplotlib.pyplot.figure(dpi=dpi, figsize=figsize)
    gs = GridSpec(rows, cols, **kwd)
    return fig, gs


def legend_patches(labels, colors):
    """
    .. versionadded:: 0.3.1

    Makes handles (using matplotlib Patch) that can be passed to the legend
    method of a matplotlib axes instance

    Arguments:
        labels (iterable): iterable that yields a label
        colors (iterable): iterable that yields a valid matplotlib color

    Returns:
        list: list of patches that can be passed to the `handles` parameter in
        the `ax.legend` method
    """
    return [
        mpatches.Patch(color=color, label=label)
        for label, color in zip(labels, colors)
    ]


__all__ = ['get_grid_figure', 'get_single_figure', 'legend_patches']



"""
This script parses results of SNPs analysis from any tool for SNP calling [#]_
and integrates them into a format that can be later used for other scripts in
the pipeline.

It integrates coverage and expected number of syn/nonsyn change and taxonomy
from a GFF file, SNP data from a VCF file.

.. note::

    The script accept gzipped VCF files

.. [#] GATK pipeline was tested, but it is possible to use samtools and
    bcftools

Changes
*******

.. versionchanged:: 0.2.1
    added *-s* option for VCF files generated using bcftools

.. versionchanged:: 0.1.16
    reworkked internals and removed SNPDat, syn/nonsyn evaluation is internal

.. versionchanged:: 0.1.13
    reworked the internals and the classes used, including options -m and -s


"""

from __future__ import division
from builtins import zip
import HTSeq
import logging
import argparse
from . import utils
from ..io import gff, compressed_handle, fasta
from .. import logger
from ..snps.classes import GeneSNP, SNPType
import pickle


LOG = logging.getLogger(__name__)


def set_parser():
    """
    Sets command line arguments parser
    """
    parser = argparse.ArgumentParser(
        description='SNPs analysis, requires a vcf file and SNPDat results',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '-o',
        '--output-file',
        default='snp_data.pickle',
        type=argparse.FileType('wb'),
        help='Ouput file'
    )
    parser.add_argument(
        '-q',
        '--min-qual',
        default=30,
        type=int,
        action='store',
        help='Minimum SNP quality (Phred score)'
    )
    parser.add_argument(
        '-f',
        '--min-freq',
        default=0.01,
        type=float,
        action='store',
        help='Minimum allele frequency'
    )
    parser.add_argument(
        '-r',
        '--min-reads',
        default=4,
        type=int,
        action='store',
        help='Minimum number of reads to accept the SNP'
    )
    parser.add_argument(
        '-g',
        '--gff-file',
        required=True,
        type=argparse.FileType('rb'),
        action='store',
        help='GFF file with annotations'
    )
    parser.add_argument(
        '-p',
        '--vcf-file',
        required=True,
        type=argparse.FileType('r'),
        action='store',
        help='Merged VCF file'
    )
    parser.add_argument(
        '-a',
        '--reference',
        required=True,
        type=argparse.FileType('rb'),
        help='Fasta file with the GFF Reference'
    )
    parser.add_argument(
        '-m',
        '--samples-id',
        action='append',
        required=True,
        type=str,
        help='the ids of the samples used in the analysis'
    )
    parser.add_argument(
        '-c',
        '--cov-suff',
        action='store',
        default='_cov',
        help="Per sample coverage suffix in the GFF"
    )
    parser.add_argument(
        '-s',
        '--bcftools-vcf',
        action='store_true',
        default=False,
        help="bcftools call was used to produce the VCF file"
    )

    utils.add_basic_options(parser, manual=__doc__)

    return parser


def init_count_set(annotations):
    LOG.info("Init data structures")

    samples = list(annotations[0].sample_coverage.keys())

    snp_data = dict(
        (sample, {}) for sample in samples
    )

    for annotation in annotations:

        taxon_id = annotation.taxon_id

        uid = annotation.uid

        sample_coverage = annotation.sample_coverage

        for sample in sample_coverage:
            snp_data[sample][uid] = GeneSNP(
                uid=uid,
                gene_id=annotation.gene_id,
                taxon_id=taxon_id,
                exp_syn=annotation.exp_syn,
                exp_nonsyn=annotation.exp_nonsyn,
                coverage=sample_coverage[sample],
            )

    return snp_data


def check_snp_in_set(samples, snp_data, pos, change, annotations, seq):
    """
    Used by :func:`parse_vcf` to check if a SNP

    :param iterable samples: list of samples that contain the SNP
    :param dict snp_data: dictionary from :func:`init_count_set` with per
        sample SNPs information
    """

    for annotation in annotations:
        if pos not in annotation:
            continue

        if annotation.is_syn(seq, pos, change, strict=False):
            snp_type = SNPType.syn
        else:
            snp_type = SNPType.nonsyn

        uid = annotation.uid
        rel_pos = annotation.get_relative_pos(pos)

        for sample in samples:
            snp_data[sample][uid].add_snp(rel_pos, change, snp_type=snp_type)


def parse_vcf(vcf_file, snp_data, min_reads, min_af, min_qual, annotations,
              seqs, options, line_num=100000):
    """
    Parse VCF file counts synonymous and non-synonymous SNPs

    :param file vcf_file: file handle to a VCF file
    :param dict snp_data: dictionary from :func:`init_count_set` with per
        sample SNPs information
    :param int min_reads: minimum number of reads to accept a SNP
    :param float min_af: minimum allele frequency to accept a SNP
    :param int min_qual: minimum quality (Phred score) to accept a SNP
    :param dict annotations: annotations grouped by their reference sequence
    :param dict seqs: reference sequences
    :param int line_num: the interval in number of lines at which progress
        will be printed
    """
    vcf_handle = HTSeq.VCF_Reader(compressed_handle(vcf_file))

    vcf_handle.parse_meta()
    vcf_handle.make_info_dict()

    # total number of SNPs accepted
    count_tot = 0
    # number of SNPs skipped for low depth
    skip_dp = 0
    # number of SNPs skipped for low allele frequency
    skip_af = 0
    # number of SNPs skipped for low quality
    skip_qual = 0
    # indels
    skip_indels = 0

    for vcf_record in vcf_handle:
        # the SNP is a sequence with no annotations
        if vcf_record.chrom not in annotations:
            continue

        if float(vcf_record.qual) < min_qual:
            # low quality SNP
            skip_qual += 1
            continue

        # unpack info records (needed for vcf_record.info to be a dictionary)
        vcf_record.unpack_info(vcf_handle.infodict)

        if vcf_record.info['INDEL']:
            skip_indels += 1
            continue

        if not isinstance(vcf_record.info['DP'], int):
            LOG.warning(vcf_record.info['DP'])

        if vcf_record.info['DP'] < min_reads:
            # not enough reads (depth) for the SNP
            skip_dp += 1
            continue

        # Samtools mpileup -> bcftools call doesn't output the allele freq.
        # it can be calculated with AC/AN for each ALT nucleotide
        # checked on bfctools (roh command) manual
        # https://samtools.github.io/bcftools/bcftools.html
        try:
            allele_freqs = vcf_record.info['AF']
        except KeyError:
            if isinstance(vcf_record.info['AC'], list):
                allele_freqs = [
                    AC / vcf_record.info['AN'] for AC in vcf_record.info['AC']
                ]
            else:
                allele_freqs = vcf_record.info['AC'] / vcf_record.info['AN']

        # if the allele frequency is a single value, make it a list, so
        # the iteration below works anyway
        if isinstance(allele_freqs, float):
            allele_freqs = [allele_freqs]

        # alt is the nucleotidic change
        iter_data = zip(allele_freqs, vcf_record.alt)
        for alt_index, (allele_freq, change) in enumerate(iter_data):
            if allele_freq < min_af:
                # the allele frequency for the SNP is too low, it'll be
                # skipped
                skip_af += 1
                continue

            # the samples that contain the SNP is a string separated by '-'
            if options.bcftools_vcf:
                samples = set()
                for sample_id, sample_info in vcf_record.samples.items():
                    # prepare the genotype list, to make the comparison easier
                    # the genotype separator to '/' only, to use only one
                    # type of split
                    sample_info_gt = sample_info['GT'].replace('|', '/')
                    sample_info_gt = sample_info_gt.split('/')
                    for genotype in sample_info_gt:
                        if genotype == '.':
                            continue
                        if int(genotype) == (alt_index + 1):
                            samples.add(sample_id)
            else:
                samples = [
                    sample
                    for sample in vcf_record.info['set'].split('-')
                ]
            check_snp_in_set(
                samples,
                snp_data,
                vcf_record.pos.start,
                change,
                annotations[vcf_record.chrom],
                seqs[vcf_record.chrom]
            )
            # increase the total number of snps available
            count_tot += 1

        if vcf_handle.line_no % line_num == 0:
            LOG.info(
                "Line %d, SNPs passed %d; skipped for: qual %d, " +
                "depth %d, freq %d, indels %d",
                vcf_handle.line_no, count_tot, skip_qual, skip_dp, skip_af,
                skip_indels
            )


def save_data(output_file, snp_data):
    """
    Pickle data structures to the disk.

    :param str output_file: base name for pickle files
    :param dict snp_data: dictionary from :func:`init_count_set` with per
        sample SNPs information
    """

    LOG.info("Saving sample SNPs to %s", output_file)
    pickle.dump(snp_data, output_file, -1)


def main():
    "Main function"
    options = set_parser().parse_args()

    # configs log and set log level
    logger.config_log(options.verbose)

    seqs = dict(fasta.load_fasta(options.reference))

    # Loads them as list because it's easier to init the data structure
    annotations = list(gff.parse_gff(options.gff_file))

    if len(annotations[0].sample_coverage) != len(options.samples_id):
        utils.exit_script(
            "Coverage information was not found for all samples", 2
        )

    snp_data = init_count_set(annotations)

    # Group annotations by their reference sequence
    annotations = gff.group_annotations(
        annotations,
        key_func=lambda x: x.seq_id
    )

    parse_vcf(
        options.vcf_file,
        snp_data,
        options.min_reads,
        options.min_freq,
        options.min_qual,
        annotations,
        seqs,
        options
    )

    save_data(options.output_file, snp_data)


if __name__ == '__main__':
    main()



"""
.. versionadded:: 0.3.0

Scripts that includes some functionality to help use FASTA files with the
framework

split command
*************

Used to split a fasta file into smaller fragments

translate command
*****************

Used to translate nucleotide sequences into amino acids.

uid command
***********

Used to change a FASTA file headers to a unique ID. A table (tab separated)
with the changes made can be kept, using the *--table* option.

Changes
*******

.. versionadded:: 0.3.0

.. versionchanged:: 0.3.1
    added *translate* and *uid* command

.. versionchanged:: 0.3.4
    ported to *click*

"""

from __future__ import division
from builtins import range
import logging
from uuid import uuid4
import click
from tqdm import tqdm
import mgkit
from . import utils
from mgkit.io import fasta
from ..utils import trans_tables
from ..utils.sequence import translate_sequence

LOG = logging.getLogger(__name__)


@click.group()
@click.version_option()
@utils.cite_option
def main():
    "Main function"
    pass


@main.command('split', help="""Splits a FASTA file [fasta-file] in a number of
              fragments""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-p', '--prefix', default='split', show_default=True,
              help='Prefix for the file name in output')
@click.option('-n', '--number', type=click.INT, default=10, show_default=True,
              help='Number of chunks into which split the FASTA file')
@click.option('-z', '--gzip', is_flag=True, default=False,
              help='gzip output files')
@click.argument('fasta-file', type=click.File('rb'), default='-')
def split_command(verbose, prefix, number, gzip, fasta_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)
    LOG.info(
        "Splitting FASTA into %d chunks with prefix (%s)",
        number,
        prefix
    )

    name_mask = "%s-{0:05}.fa" % prefix
    if gzip:
        name_mask += '.gz'
        LOG.info("Output files will be compressed (gzip)")

    fasta.split_fasta_file(
        fasta_file,
        name_mask,
        num_files=number
    )


def load_trans_table(table_name):
    "Loads translation table "
    return getattr(trans_tables, table_name.upper())


def translate_seq(name, seq, trans_table):
    "Tranlates sequence into the 6 frames"
    header = "{0}-{1}{2}"
    for start in range(3):
        yield header.format(name, 'f', start), translate_sequence(seq, start, trans_table, False)
        yield header.format(name, 'r', start), translate_sequence(seq, start, trans_table, True)


@main.command('translate', help="""Translate FASTA file [fasta-file] in all 6
              frames to [output-file]""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-t', '--trans-table', default='universal', show_default=True,
              type=click.Choice([table_name.lower() for table_name in dir(trans_tables) if not table_name.startswith('_')]),
              help='translation table')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('fasta-file', type=click.File('rb'), default='-')
@click.argument('output-file', type=click.File('wb'), default='-')
def translate_command(verbose, trans_table, progress, fasta_file, output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)
    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    trans_table = load_trans_table(trans_table)

    iterator = fasta.load_fasta(fasta_file)

    iterator = tqdm(iterator)

    for name, seq in iterator:
        for new_header, new_seq in translate_seq(name, seq, trans_table):
            fasta.write_fasta_sequence(output_file, new_header, new_seq)


@main.command('uid', help="""Changes each header of a FASTA file [file-file] to
              a uid (unique ID)""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-t', '--table', default=None, type=click.File('wb'),
              help='Filename of a table to record the changes (by default discards it)')
@click.argument('fasta-file', type=click.File('rb'), default='-')
@click.argument('output-file', type=click.File('wb'), default='-')
def uid_command(verbose, table, fasta_file, output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)
    if table is not None:
        LOG.info(
            'Writing Table to file (%s)',
            getattr(table, 'name', repr(table))
        )

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    for name, seq in fasta.load_fasta(fasta_file):
        uid = str(uuid4())
        if table is not None:
            table.write("{}\t{}\n".format(uid, name).encode('ascii'))

        fasta.write_fasta_sequence(output_file, uid, seq)



"""
Scaling functions for counts
"""

from __future__ import division
from scipy import stats
import numpy
import pandas


def scale_factor_deseq(dataframe):
    """
    .. versionadded:: 0.1.13

    Returns the scale factor according to he deseq paper. The columns of the
    dataframe are the samples.

    size factor :math:`\\hat{s}_{j}` for sample *j* (from DESeq paper).

    .. math::

        \\hat{s}_{j} = median_{i} (
        \\frac
            {k_{ij}}
            {
                \\left (
                \\prod_{v=1}^{m}
                    k_{iv}
                \\right )^{1/m}
           }
        )

    """
    # calc the genes geometric mean over all samples
    gmean = dataframe.apply(stats.gmean, axis=1)
    # keeps only the genes whose geometric mean is > 0
    gmean = gmean[gmean > 0]

    sample_factors = {}

    # calc the scaling factor for each sample
    for sample, genes in dataframe.items():

        scale_factor = numpy.median(genes.loc[gmean.index] / gmean)

        sample_factors[sample] = scale_factor

    return pandas.Series(sample_factors)


def scale_deseq(dataframe):
    """
    .. versionadded:: 0.1.13

    Scale a dataframe using the deseq scaling. Uses :func:`scale_factor_deseq`
    """
    scale_factors = scale_factor_deseq(dataframe)

    return dataframe / scale_factors


def scale_rpkm(dataframe, gene_len):
    """
    .. versionadded:: 0.1.14

    Perform an RPKM scaling of the pandas dataframe/series supplied using the
    *gene_len* series containing the gene sizes for all elements of *dataframe*

    .. math::

        RPKM =\\frac {10^{9} \\cdot C} {N \\cdot L}

    """

    gene_len = gene_len[dataframe.index]
    tot_reads = dataframe.sum().sum()

    return (10 ** 9) * dataframe.div(gene_len * tot_reads, axis='index')



"""
Module containing classes and functions to deal with CaZy data
"""

import logging

CAZY_FAMILIES = {
    'GH': 'Glycoside Hydrolase',
    'GT': 'GlycosylTransferase',
    'PL': 'Polysaccharide Lyase',
    'CE': 'Carbohydrate Esterase',
    'CBM': 'Carbohydrate-Binding Module'
}
"CaZy families"

LOG = logging.getLogger(__name__)



import pytest
from conftest import skip_no_connection
import numpy
from mgkit.snps.classes import GeneSNP, SNPType
from mgkit.snps.funcs import combine_sample_snps, flat_sample_snps
from mgkit.snps.conv_func import get_rank_dataframe, get_full_dataframe

import numpy


def test_genesyn_calc_ratio1():
    # syn and nonsyn > 0
    gs = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6,
    )
    gs.add_snp(1, 'A', SNPType.syn)
    gs.add_snp(1, 'A', SNPType.syn)
    gs.add_snp(1, 'A', SNPType.nonsyn)
    gs.add_snp(1, 'A', SNPType.nonsyn)
    gs.add_snp(1, 'A', SNPType.nonsyn)
    assert gs.calc_ratio() == 1.0


def test_genesyn_calc_ratio2():
    # syn = nonsyn = 0, haplotypes=True
    gs = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6,
    )
    assert gs.calc_ratio(haplotypes=True) == 0.0


def test_genesyn_add1():
    # checks if the conserved gene_id and taxon_id is from the first instance
    gs1 = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6
    )
    gs2 = GeneSNP(
        exp_syn=4,
        exp_nonsyn=6,
        coverage=4
    )
    gs1.add(gs2)
    assert (gs1.gene_id, gs1.taxon_id) == ('A', 1)


def test_genesyn_add2():
    # checks if the values (syn, nonsyn, exp_syn, exp_nonsyn) are added
    gs1 = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6
    )
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.nonsyn)
    gs2 = GeneSNP(
        exp_syn=3,
        exp_nonsyn=5,
        coverage=4
    )
    gs2.add_snp(1, 'A', SNPType.syn)
    gs2.add_snp(1, 'A', SNPType.nonsyn)
    gs2.add_snp(1, 'A', SNPType.nonsyn)

    gs1.add(gs2)
    assert (gs1.exp_syn, gs1.exp_nonsyn, gs1.syn, gs1.nonsyn) == (7, 11, 4, 3)


def test_genesyn_add3():
    # checks if the values if coverage is correctly added
    # coverage in gs2 is not None, in gs1 is None
    gs1 = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6
    )
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.nonsyn)
    gs2 = GeneSNP(
        exp_syn=3,
        exp_nonsyn=5,
        coverage=5
    )
    gs2.add_snp(1, 'A', SNPType.syn)
    gs2.add_snp(1, 'A', SNPType.nonsyn)
    gs2.add_snp(1, 'A', SNPType.nonsyn)

    gs1.add(gs2)
    assert gs1.coverage == 5


def test_genesyn_add4():
    # checks if the values if coverage is correctly added
    # coverage in gs2 is not None, in gs1 is not None
    gs1 = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6,
        coverage=2
    )
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.nonsyn)
    gs2 = GeneSNP(
        exp_syn=3,
        exp_nonsyn=5,
        coverage=5
    )
    gs2.add_snp(1, 'A', SNPType.syn)
    gs2.add_snp(1, 'A', SNPType.nonsyn)
    gs2.add_snp(1, 'A', SNPType.nonsyn)

    gs1.add(gs2)
    assert gs1.coverage == 7


def test_genesyn_add5():
    # checks if the values of coverage is correctly added
    # coverage in gs2 is None, in gs1 is None
    gs1 = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6,
    )
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.syn)
    gs1.add_snp(1, 'A', SNPType.nonsyn)
    gs2 = GeneSNP(
        exp_syn=3,
        exp_nonsyn=5,
        coverage=None
    )
    gs2.add_snp(1, 'A', SNPType.syn)
    gs2.add_snp(1, 'A', SNPType.nonsyn)
    gs2.add_snp(1, 'A', SNPType.nonsyn)

    gs1.add(gs2)
    assert gs1.coverage is None


def test_genesyn_calc_ratio3b():
    # syn = nonsyn = 0
    # coverage is None
    gs = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6,
    )
    assert numpy.isnan(gs.calc_ratio())


def test_genesyn_calc_ratio3c():
    # syn = nonsyn = 0
    gs = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6,
    )
    assert numpy.isnan(gs.calc_ratio())


def test_genesyn_calc_ratio3d():
    # syn = nonsyn = 0, haplotypes = True
    gs = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6,
    )
    assert gs.calc_ratio(haplotypes=True) == 0


def test_genesyn_calc_ratio4a():
    # nonsyn > 0, syn = 0
    # flag_value=True
    gs = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6,
    )
    gs.add_snp(1, 'A', SNPType.nonsyn)
    assert gs.calc_ratio_flag() == -1


def test_genesyn_calc_ratio4b():
    # nonsyn = 0, syn > 0
    # flag_value=True
    gs = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6,
    )
    gs.add_snp(1, 'A', SNPType.syn)
    assert gs.calc_ratio_flag() == -2


def test_genesyn_calc_ratio4c():
    # nonsyn = 0, syn > 0
    # flag_value=True
    gs = GeneSNP(
        gene_id='A',
        taxon_id=1,
        exp_syn=4,
        exp_nonsyn=6,
    )
    assert gs.calc_ratio_flag() == -3


SNP_DATA = {
    'sample1': {
        'gene1': GeneSNP(
            gene_id='gene1',
            taxon_id=839,  # prevotella ruminicola
            exp_syn=6,
            exp_nonsyn=4,
            snps=[
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.nonsyn),
                (1, 'A', SNPType.nonsyn),
            ]
        ),  # pN/pS = 1.0
        'gene2': GeneSNP(
            gene_id='gene2',
            taxon_id=838,  # prevotella genus
            exp_syn=3,
            exp_nonsyn=4,
            snps=[
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.nonsyn),
                (1, 'A', SNPType.nonsyn),
            ]
        )  # pN/pS = 2.0
    },
    'sample2': {
        'gene1': GeneSNP(
            gene_id='gene1',
            taxon_id=839,  # prevotella ruminicola
            exp_syn=3,
            exp_nonsyn=4,
            snps=[
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.nonsyn),
                (1, 'A', SNPType.nonsyn),
            ]
        ),  # pN/pS = 1.0
        'gene2': GeneSNP(
            gene_id='gene2',
            taxon_id=838,  # prevotella genus
            exp_syn=6,
            exp_nonsyn=4,
            snps=[
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.nonsyn),
                (1, 'A', SNPType.nonsyn),
            ]
        ),  # pN/pS = 2.0
        'gene3': GeneSNP(
            gene_id='gene3',
            taxon_id=838,  # prevotella genus
            exp_syn=3,
            exp_nonsyn=4,
            snps=[
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.nonsyn),
                (1, 'A', SNPType.nonsyn),
            ]
        )  # pN/pS = 2.0
    }

}


SNP_DATA2 = {
    'sample1': {
        'gene1': GeneSNP(
            gene_id='gene1',
            taxon_id=839,  # prevotella ruminicola
            exp_syn=6,
            exp_nonsyn=4,
            coverage=4,
            snps=[
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.nonsyn),
                (1, 'A', SNPType.nonsyn),
            ]
        ),  # pN/pS = 1.0
        'gene2': GeneSNP(
            gene_id='gene2',
            taxon_id=838,  # prevotella genus
            exp_syn=3,
            exp_nonsyn=4,
            coverage=4,
            snps=[
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.nonsyn),
                (1, 'A', SNPType.nonsyn),
            ]
        )  # pN/pS = 2.0
    },
    'sample2': {
        'gene1': GeneSNP(
            gene_id='gene1',
            taxon_id=839,  # prevotella ruminicola
            exp_syn=3,
            exp_nonsyn=4,
            coverage=4,
            snps=[
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.nonsyn),
                (1, 'A', SNPType.nonsyn),
            ]
        ),  # pN/pS = 1.0
        'gene2': GeneSNP(
            gene_id='gene2',
            taxon_id=838,  # prevotella genus
            exp_syn=6,
            exp_nonsyn=4,
            coverage=3,
            snps=[
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.syn),
                (1, 'A', SNPType.nonsyn),
                (1, 'A', SNPType.nonsyn),
            ]
        )  # pN/pS = 2.0
    }

}


def test_flat_sample_snps1():
    data = flat_sample_snps(SNP_DATA2, 4)
    assert (
            data['all_samples']['gene1'].syn,
            data['all_samples']['gene2'].syn
        ) == (7, 3)


def test_flat_sample_snps2():
    data = flat_sample_snps(SNP_DATA2, 3)
    assert (
            data['all_samples']['gene1'].syn,
            data['all_samples']['gene2'].syn
        ) == (7, 8)


def test_combine_sample_min_num1():
    # check if min_num is working
    df = combine_sample_snps(SNP_DATA, 1, [lambda x: x])
    df.shape == (3, 2)


def test_combine_sample_snps_min_num2():
    # check if min_num is working
    df = combine_sample_snps(SNP_DATA, 2, [lambda x: x])
    assert df.shape == (2, 2)


def test_combine_sample_snps_values():
    # check if values are correct is working
    df = combine_sample_snps(SNP_DATA, 1, [lambda x: x])
    assert (df.min().min(), df.max().max()) == (0.5, 1.)


def test_combine_sample_snps_index1():
    # check index_type
    df = combine_sample_snps(SNP_DATA, 1, [lambda x: x])
    assert df.index.tolist() == [('gene1', 839), ('gene2', 838), ('gene3', 838)]


def test_combine_sample_snps_index2():
    # check index_type
    df = combine_sample_snps(
        SNP_DATA,
        1,
        [lambda x: x],
        index_type='gene'
    )
    # Can't expect the list to be sorted
    assert sorted(df.index.tolist()) == ['gene1', 'gene2', 'gene3']


def test_combine_sample_snps_index3():
    # check index_type
    df = combine_sample_snps(
        SNP_DATA,
        1,
        [lambda x: x],
        index_type='taxon'
    )
    assert df.index.tolist() == [838, 839]


@skip_no_connection
def test_get_rank_dataframe(ncbi_taxonomy):
    df = get_rank_dataframe(SNP_DATA2, ncbi_taxonomy, min_num=1, rank='genus')
    assert (df.loc[838, 'sample1'] == .75) and (df.loc[838, 'sample2'] == .375)


@skip_no_connection
def test_get_full_dataframe(ncbi_taxonomy):
    df = get_full_dataframe(SNP_DATA2, ncbi_taxonomy, min_num=1)
    assert df.loc['gene2'].loc[838, 'sample1'] == .5
    assert numpy.isnan(df.loc[('gene2', 838), 'sample2'])
    assert df.loc[('gene1', 839), 'sample1'] == 1.
    assert df.loc[('gene1', 839), 'sample2'] == .375



"""
.. versionadded:: 0.1.12

Graph module
"""
import logging
from future.utils import viewitems
from xml.etree import ElementTree
import itertools
from . import DependencyError

try:
    import networkx as nx
except ImportError:
    raise DependencyError('networkx')

LOG = logging.getLogger(__name__)


def build_graph(id_links, name, edge_type='', weight=0.5):
    """
    .. versionadded:: 0.1.12

    Builds a networkx graph from a dictionary of nodes, as outputted by
    :meth:`mgkit.kegg.KeggClientRest.get_pathway_links`. The graph is
    undirected, and all edges weight are the same.

    Arguments:
        id_links (dict): dictionary with the links
        name (str): name of the graph
        edge_type (str): an optional name for the `edge_type` attribute
            set for each edge
        weight (float): the weight assigned to each edge in the graph

    Returns:
        graph: an instance of :class:`networkx.Graph`
    """
    g = nx.Graph()
    g.name = name

    for id1, id2s in viewitems(id_links):
        g.add_node(id1, id=id1)
        for id2 in id2s:
            g.add_node(id2, id=id2)
            g.add_edge(id1, id2, edge_type=edge_type, weight=weight)

    return g


def build_weighted_graph(id_links, name, weights, edge_type=''):
    """
    .. versionadded:: 0.1.14

    Builds a networkx graph from a dictionary of nodes, as outputted by
    :meth:`mgkit.kegg.KeggClientRest.get_pathway_links`. The graph is
    undirected, and all edges weight are the same.

    Arguments:
        id_links (dict): dictionary with the links
        name (str): name of the graph
        edge_type (str): an optional name for the `edge_type` attribute
            set for each edge
        weight (float): the weight assigned to each edge in the graph

    Returns:
        graph: an instance of :class:`networkx.Graph`
    """
    g = nx.Graph()
    g.name = name

    for id1, id2s in id_links:
        g.add_node(id1, id=id1)
        for id2 in id2s:
            g.add_node(id2, id=id2)
            try:
                weight = weights[(id1, id2)]
            except KeyError:
                weight = weights.get(id1, 0.5)
            g.add_edge(id1, id2, edge_type=edge_type, weight=float(weight))

    return g


def rename_graph_nodes(graph, name_func=None, exclude_ids=None):
    new_graph = nx.Graph()

    if exclude_ids is None:
        exclude_ids = set()

    for node, data in graph.nodes_iter(data=True):
        new_graph.add_node(
            node if node in exclude_ids else name_func(node),
            **data
        )

    for node1, node2, data in graph.edges_iter(data=True):
        new_graph.add_edge(
            node1 if node1 in exclude_ids else name_func(node1),
            node2 if node2 in exclude_ids else name_func(node2),
            **data
        )

    return new_graph


def copy_nodes(g, graph1, name=None, id_attr=None, **kwd):
    """
    .. versionadded:: 0.1.12

    Used by :func:`link_nodes` to copy nodes
    """

    if name is None:
        name = graph1.name

    if id_attr is None:
        id_attr = 'id'

    for node, data in graph1.nodes_iter(data=True):
        data = data.copy()
        data.update(kwd)
        g.add_node(
            "{0}_{1}".format(name, data[id_attr]),
            **data
        )


def copy_edges(g, graph1, name=None, **kwd):
    """
    .. versionadded:: 0.1.12

    Used by :func:`link_nodes` to copy edges
    """

    if name is None:
        name = graph1.name

    for node1, node2, data in graph1.edges_iter(data=True):
        data = data.copy()
        data.update(kwd)
        g.add_edge(
            "{0}_{1}".format(name, node1),
            "{0}_{1}".format(name, node2),
            **data
        )


def link_nodes(g, graph1, graph2, id_filter, link_type, weight):
    """
    .. versionadded:: 0.1.12

    Used by :func:`link_graph` to link nodes with the same *id*
    """
    nodes1 = set(
        data['id']
        for node, data in graph1.nodes_iter(data=True)
        if id_filter(data['id'])
    )
    nodes2 = set(
        data['id']
        for node, data in graph2.nodes_iter(data=True)
        if id_filter(data['id'])
    )

    for node_id in (nodes1 & nodes2):
        g.node["{0}_{1}".format(graph1.name, node_id)]['alpha'] = 0.5
        g.node["{0}_{1}".format(graph2.name, node_id)]['alpha'] = 0.5
        g.add_edge(
            "{0}_{1}".format(graph1.name, node_id),
            "{0}_{1}".format(graph2.name, node_id),
            edge_type=link_type,
            weight=weight
        )


EDGE_LINKS = [
    (lambda x: x.startswith('C'), 'CP_LINK', 0.0),
    (lambda x: x.startswith('K'), 'KO_LINK', 0.0)
]
"Sample edge_links for :func:`link_graph`"


def link_graph(graphs, edge_links):
    """
    .. versionadded:: 0.1.12

    Link nodes of a set of graphs using the specifics in edge_links.
    The resulting graph nodes are renamed, and the nodes that are shared
    between the graphs linked.

    Arguments:
        graphs: iterable of graphs
        edge_links: iterable with function, edge_type and weight for the
            links between graphs

    Returns:
        graph: an instance of :class:`networkx.Graph`
    """
    g = nx.Graph()

    for graph in graphs:
        copy_nodes(g, graph)
        copy_edges(g, graph)

    for graph1, graph2 in itertools.combinations(graphs, r=2):
        for id_filter, link_type, weight in edge_links:
            link_nodes(g, graph1, graph2, id_filter, link_type, weight)

    return g


def filter_graph(graph, id_list, filter_func=lambda x: x.startswith('K')):
    """
    .. versionadded:: 0.1.12

    Filter a graph based on the `id_list` provided and the filter function
    used to test the id attribute of each node.

    A node is removed if `filter_func` returns True on a node and its id
    attribute is not in `id_list`

    Arguments:
        graph: the graph to filter
        id_list (iterable): the list of nodes that are to remain in the
            graph
        filter_func (func): function which accept a single parameter and
            return a boolean

    Returns:
        graph: an instance of :class:`networkx.Graph`
    """
    graph = graph.copy()
    nodes = [
        node
        for node, data in graph.nodes_iter(data=True)
        if filter_func(data['id']) and (data['id'] not in id_list)
    ]

    graph.remove_nodes_from(nodes)

    graph.remove_nodes_from(nx.isolates(graph))

    return graph


def annotate_graph_nodes(graph, attr, id_map, default=None, conv=None):
    """
    .. versionadded:: 0.1.12

    .. versionchanged:: 0.4.0
        added *conv* parameter and reworked internals

    Add/Changes nodes attribute `attr` using a dictionary of ids->values.

    .. note::

        If the id is not found in `id_map`:

        * default is None: no value added for that node
        * default is not None: the node attribute will be set to `default`

    Arguments:
        graph: the graph to annotate
        attr (str): the attribute to annotate
        id_map (dict): the dictionary with the values for each node
        default: the value used in case an `id` is not found in `id_map`, if
            None, the attribute is not set for missing values
        conv (func): function to convert the value to another type
    """
    for node_id in graph.nodes():
        value = id_map.get(node_id, default)
        if value is None:
            continue
        if conv is not None:
            value = conv(value)
        graph.nodes[node_id][attr] = value


def from_kgml(entry, graph=None, rn_ids=None):
    """
    .. versionadded:: 0.3.1

    Given a KGML file (as string), representing a pathway in Kegg, returns a
    networkx DiGraph, using reaction directionality included in the KGML. If a
    reaction is reversible, 2 edges (from and to) for each compound/reaction
    pair are added, giving the bidirectionality.

    .. note::

        substrate and products included in a KGML don't represent the complete
        reaction, excluding in general cofactors or more general terms.
        Those can be added using :func:`add_module_compounds`, which may be
        more useful when used with a restricted number of reactions (e.g.
        a module)

    Arguments:
        entry (str): KGML file as a string, or anything that can be passed to
            ElementTree
        graph (graph): an instance of a networkx DiGraph if the network is to
            be updated with a new KGML, if `None` a new one is created
        rn_ids (set): a set/list of reaction IDs that are to be included, if
            `None` all reactions are used

    Returns:
        graph: a networkx DiGraph with the reaction/compounds
    """
    if graph is None:
        graph = nx.DiGraph()

    for entry in ElementTree.fromstring(entry).findall('reaction'):
        # the "reaction" defined is equivalent to a EC number, meaning multiple
        # reactions IDs in kegg may belong to it. They are stored in the name
        # attribute, separated by space
        reactions = entry.attrib['name'].split(' ')

        for reaction in reactions:
            # Each reaction ID is (as usual) prepended by the type "rn", which
            # we don't need
            reaction = reaction.split(':')[1]
            if (rn_ids is not None) and (reaction not in rn_ids):
                continue
            # definition of the reaction direction, by manual either reversible
            # or irreversible
            if entry.attrib['type'] == 'irreversible':
                reversible = False
            else:
                reversible = True

            substrates = []
            products = []

            graph.add_node(
                reaction,
                node_type='reaction',
                reaction_type='reversible' if reversible else 'irreversible'
            )

            # separating substrate and products from the compounds listed in
            # the reaction definition
            for compound in entry:
                cpd_id = compound.attrib['name'].split(':')[1]
                if compound.tag == 'substrate':
                    substrates.append(cpd_id)
                else:
                    products.append(cpd_id)

            for substrate in substrates:
                if substrate not in graph:
                    graph.add_node(substrate, node_type='substrate')
                else:
                    # cases where the substrate is product of other reactions
                    if graph.node[substrate]['node_type'] != 'substrate':
                        graph.node[substrate]['node_type'] = 'mixed'

                graph.add_edge(
                    substrate,
                    reaction,
                    reaction_type='reversible' if reversible else 'irreversible'
                )
                # if reversible add the reciprocal edge
                if reversible:
                    graph.add_edge(
                        reaction,
                        substrate,
                        reaction_type='reversible' if reversible else 'irreversible'
                    )

            for product in products:
                if product not in graph:
                    graph.add_node(product, node_type='product')
                else:
                    # cases where the product is product of other reactions
                    if graph.node[product]['node_type'] != 'product':
                        graph.node[product]['node_type'] = 'mixed'

                graph.add_edge(
                    reaction,
                    product,
                    reaction_type='reversible' if reversible else 'irreversible'
                )
                # if reversible add the reciprocal edge
                if reversible:
                    graph.add_edge(
                        product,
                        reaction,
                        reaction_type='reversible' if reversible else 'irreversible'
                    )

    return graph


def add_module_compounds(graph, rn_defs):
    """
    .. versionadded:: 0.3.1

    Modify in-place a graph, by adding additional compounds from a dictionary
    of definitions. It uses the reversible/irreversible information for each
    reaction to add the correct number of edges to the graph.

    Arguments:
        graph (graph): a graph to update with additional compounds
        rn_defs (dict): a dictionary, whose keys are reactions IDs and the
            values are instances of :class:`mgkit.kegg.KeggReaction`
    """
    for rn_id, (left_cp, right_cp) in viewitems(rn_defs):

        reaction_type = graph.node[rn_id]['reaction_type']
        reversible = True if reaction_type == 'reversible' else False
        for cp_id in left_cp | right_cp:
            if ((rn_id, cp_id) in graph.edges()) or ((rn_id, cp_id) in graph.edges()):
                continue
            if cp_id not in graph:
                graph.add_node(cp_id, node_type='addition')
            if reversible:
                graph.add_edge(rn_id, cp_id, reaction_type=reaction_type)
                graph.add_edge(cp_id, rn_id, reaction_type=reaction_type)
            else:
                if cp_id in graph.predecessors(rn_id):
                    graph.add_edge(cp_id, rn_id, reaction_type=reaction_type)
                else:
                    graph.add_edge(rn_id, cp_id, reaction_type=reaction_type)


class Reaction(object):
    """
    .. versionadded:: 0.4.0

    Object used to hold information about a reaction entry in Kegg
    """
    kegg_id = None
    substrates = None
    products = None
    reversible_paths = None
    irreversible_paths = None
    orthologs = None

    def __init__(self, kegg_id, substrates, products, reversible, orthologs, pathway):
        self.kegg_id = kegg_id
        self.substrates = set(substrates)
        self.products = set(products)
        self.irreversible_paths = set()
        self.reversible_paths = set()
        if reversible:
            self.reversible_paths.add(pathway)
        else:
            self.irreversible_paths.add(pathway)
        self.orthologs = set(orthologs)

    @property
    def reversible(self):
        """
        Property that returns the reversibility of the reaction according to
        the information in the pathways. Returns True if the number of pathways
        in which the reaction was observed as reversible is greater or equal
        than the number of pathwaysin which the reaction was observerd as
        irreversible.
        """
        return len(self.reversible_paths) >= len(self.irreversible_paths)

    @property
    def pathways(self):
        """
        Set which includes all the pathways in which the reaction was found
        """
        return self.irreversible_paths | self.reversible_paths

    def update(self, other):
        """
        Updates the current instance with information from another instance.
        the underlining sets that hold the information are update with those
        from the `other` instance.

        Raises:
            ValueError: if the ID of the reaction is different
        """
        if self.kegg_id != other.kegg_id:
            raise ValueError('The reactions have different IDs')

        # case where substrates and products are inverted
        if (self.substrates == other.products) and (other.substrates == self.products):
            pass
        # case in which they are inverted but contain a subset
        elif (self.substrates & other.products) or (other.substrates & self.products):
            # checks for substrates
            if self.substrates & other.products:
                self.substrates.update(other.substrates - self.products)
            # checks for products
            elif self.products & other.substrates:
                self.products.update(other.products - self.substrates)
            # logs a warning
            else:
                LOG.warning("Unknown State %r", self)
        else:
            self.substrates.update(other.substrates)
            self.products.update(other.products)
        self.reversible_paths.update(other.reversible_paths)
        self.irreversible_paths.update(other.irreversible_paths)
        self.orthologs.update(other.orthologs)

    def cmp_compounds(self, other):
        """
        Compares the substrates and products of the current instance with those
        of another one, using information about the reversibility of the
        reaction.
        """
        if (self.substrates != other.substrates) and (self.products != other.products):
            if self.reversible:
                return (self.substrates == other.products) and (self.products == other.substrates)
            else:
                return False
        else:
            return True

    def to_nodes(self):
        """
        Returns a generator that returns the nodes associated with reaction,
        to be used in a graph, along with attributes about the type of node
        (reaction or compound).
        """
        return itertools.chain(
            [(self.kegg_id, dict(type='reaction'))],
            [(cpd, dict(type='compound')) for cpd in self.substrates],
            [(cpd, dict(type='compound')) for cpd in self.products],
        )

    def to_edges(self):
        """
        Returns a generator of edges to be used when building a graph, along
        with an attribute that specify if the reaction is reversible.
        """
        edges = [
            itertools.product(self.substrates, [self.kegg_id]),
            itertools.product([self.kegg_id], self.products)
        ]
        if self.reversible:
            edges.extend(
                [
                    itertools.product(self.products, [self.kegg_id]),
                    itertools.product([self.kegg_id], self.substrates)
                ]
            )
        return itertools.chain(*edges), dict(reversible=self.reversible)

    def to_edges_compounds(self):
        edges = list(itertools.product(self.substrates, self.products))
        if self.reversible:
            edges.extend(itertools.product(self.products, self.substrates))

        edges = [(node1, node2, self.kegg_id) for node1, node2 in edges if node1 != node2]
        # Key is used in MultiGraphs to distinguish the edges of multiple
        attr = dict(key=self.kegg_id)

        return edges, attr

    def __eq__(self, other):
        """
        Tests equality by comparing the IDs and the compounds
        """
        if (self.kegg_id == other.kegg_id) and (self.reversible == other.reversible) and (self.orthologs == other.orthologs):
            return self.cmp_compounds(other)

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "{}: s {} {} p {} - ({} - {})".format(
            self.kegg_id,
            ','.join(self.substrates),
            '<=>' if self.reversible else '=>',
            ','.join(self.products),
            ','.join(self.orthologs),
            ','.join(self.pathways)
        )

    def __str__(self):
        return repr(self)


def parse_kgml_reactions(kgml):
    """
    .. versionadded:: 0.4.0

    Parses a KGML for reactions, returning a dictionary with instances of
    :class:`Reaction` as values and the IDs as keys.

    Arguments:
        kgml (str): the KGML file content as a string (to be passed)

    Returns:
        dict: dictionary of ID->Reaction
    """
    pathway = ElementTree.fromstring(kgml)
    pathway_name = pathway.attrib['name'].replace('path:', '')
    assert len(pathway_name) == 7

    reaction_data = {}

    for entry in pathway.findall('reaction'):
        orthologs = tuple(
            sorted(
                ortholog.replace('ko:', '')
                for ortholog in pathway.find("entry/[@id='{}']".format(entry.attrib['id'])).attrib['name'].split(' ')
            )
        )
        assert all(len(x) == 6 for x in orthologs)
        # the "reaction" defined is equivalent to a EC number, meaning multiple
        # reactions IDs in kegg may belong to it. They are stored in the name
        # attribute, separated by space
        reactions = entry.attrib['name'].split(' ')
        # definition of the reaction direction, by manual either reversible
        # or irreversible
        if entry.attrib['type'] == 'irreversible':
            reversible = False
        else:
            reversible = True

        substrates = set()
        products = set()

        for compound in entry:
            cpd_ids = compound.attrib['name'].split(' ')
            cpd_ids = [cpd_id.split(':')[1].strip() for cpd_id in cpd_ids]
            assert all(len(cpd_id) == 6 for cpd_id in cpd_ids)
            for cpd_id in cpd_ids:
                if not (cpd_id.startswith('C') or cpd_id.startswith('G') or cpd_id.startswith('D')):
                    continue
            if compound.tag == 'substrate':
                substrates.update(cpd_ids)
            else:
                products.update(cpd_ids)

        assert len(substrates) + len(products) > 1

        for reaction in reactions:
            # Each reaction ID is (as usual) prepended by the type "rn", which
            # we don't need
            reaction = reaction.split(':')[1]
            assert len(reaction) == 6

            definition = Reaction(reaction, substrates, products, reversible, orthologs, pathway_name)

            try:
                reaction_data[reaction].update(definition)
            except KeyError:
                reaction_data[reaction] = definition

    return reaction_data


def merge_kgmls(kgmls):
    """
    .. versionadded:: 0.4.0

    Parses multiple KGMLs and merges the reactions from them.

    Arguments:
        kgmls (iterable): iterable of KGML files (content) to be passed to
            :func:`parse_kgml_reactions`

    Returns:
        dict: dictionary with the reactions from amm te KGML files
    """
    combined_data = {}

    for kgml in kgmls:
        data = parse_kgml_reactions(kgml)
        for kegg_id, reaction in viewitems(data):
            try:
                combined_data[kegg_id].update(reaction)
            except KeyError:
                combined_data[kegg_id] = reaction

    return combined_data



"""
Module that contains mapping operations on pandas data structures
"""
from __future__ import division
from future.utils import viewitems
from builtins import zip
import pandas


def group_dataframe_by_mapping(dataframe, mapping, root_taxon, name_dict=None):
    """
    Return a :class:`pandas.DataFrame` filtered by mapping and root taxon, the
    values for each column is averaged over all genes mapping to a category.

    :param DataFrame dataframe: DataFrame with multindex gene-root
    :param dict mapping: dictionary of category->genes
    :param str root_taxon: root taxon to group genes
    :param dict name_dict: dictionary of category->name

    :return DataFrame: DataFrame filtered
    """
    data_dict = {}
    for category, id_list in viewitems(mapping):
        category_name = category
        if name_dict is not None:
            category_name = name_dict[category_name]
        data_dict[category_name] = dataframe.loc[id_list].xs(
            root_taxon, level='root'
        ).mean()

    return pandas.DataFrame.from_dict(data_dict, orient='index')


def calc_coefficient_of_variation(dataframe):
    """
    Calculate coefficient of variation for a DataFrame. Uses formula from
    `Wikipedia <http://en.wikipedia.org/wiki/Coefficient_of_variation>`_

    The formula used is :math:`\\left (1 + \\frac {1}{4n} \\right ) * c_{v}`
    where :math:`c_{v} = \\frac {s}{\\bar{x}}`
    """

    dataframe_cv = dataframe.std(axis=1) / dataframe.mean(axis=1)
    coeff = 1 + (1.0 / (4 * dataframe.count(axis=1)))

    return coeff * dataframe_cv


def make_stat_table(dataframes, roots):
    """
    Produces a :class:`pandas.DataFrame` that summarise the supplied
    DataFrames. The stats include mean, stdev and coefficient of variation for
    each root taxon.

    :param iterable dataframes: iterable of DataFrame instances
    :param iterable roots: list of root taxa to which each table belongs

    :return DataFrame: returns a DataFrame instance
    """

    index = []

    data = {}

    for dataframe, root in zip(dataframes, roots):
        index.append((root, 'mean'))
        data[(root, 'mean')] = dataframe.mean(axis=1)
        index.append((root, 'stdev'))
        data[(root, 'stdev')] = dataframe.std(axis=1)
        index.append((root, 'c. var'))
        data[(root, 'c. var')] = calc_coefficient_of_variation(dataframe)

    index = pandas.MultiIndex.from_tuples(sorted(index),
                                          names=('root', 'value'))

    return pandas.DataFrame(data, columns=index)


def concatenate_and_rename_tables(dataframes, roots):
    """
    Concatenates a list of :class:`pandas.DataFrame` instances and renames the
    columns prepending a string to each column in each table from a list of
    prefixes.

    :param iterable dataframes: iterable of DataFrame instances
    :param iterable roots: list of prefixes to append to the column names of
        each DataFrame

    :return DataFrame: returns a DataFrame instance

    .. todo::

        * move to pandas_utils?
    """
    renamed = []

    for dataframe, root in zip(dataframes, roots):
        index = dict((column, root + column) for column in dataframe.columns)
        renamed.append(dataframe.rename(columns=index))

    return pandas.concat(renamed, axis=1)



import pytest
import functools

from mgkit.io.gff import Annotation
from mgkit.filter.gff import choose_annotation, filter_annotations, \
    filter_base, filter_len, filter_attr_num, filter_attr_str, \
    filter_base_num, filter_attr_num_s


def test_choose_annotation_contained1():
    # same size, better score
    a = Annotation(start=1, end=10, bitscore=10)
    b = Annotation(start=1, end=10, bitscore=15)
    assert choose_annotation(a, b) == a


def test_choose_annotation_contained2():
    # same score, longer
    a = Annotation(start=1, end=10, bitscore=10)
    b = Annotation(start=5, end=10, bitscore=10)
    assert choose_annotation(a, b) == b


def test_choose_annotation_overlap1():
    # overlapping, better score
    a = Annotation(start=1, end=10, bitscore=10)
    b = Annotation(start=5, end=11, bitscore=15)
    assert choose_annotation(a, b, overlap=4) == a


def test_choose_annotation_overlap2():
    # overlapping, same score, choose longer
    a = Annotation(start=1, end=11, bitscore=10)
    b = Annotation(start=5, end=11, bitscore=10)
    assert choose_annotation(a, b, overlap=4) == b


def test_choose_annotation_none1():
    # no overlap
    a = Annotation(start=1, end=10, bitscore=10)
    b = Annotation(start=11, end=12, bitscore=10)
    assert choose_annotation(a, b, overlap=4) is None


def test_choose_annotation_none2():
    # overlap lower than the limits
    a = Annotation(start=1, end=10, bitscore=10)
    b = Annotation(start=9, end=12, bitscore=10)
    assert choose_annotation(a, b, overlap=2) is None


def test_filter_annotations1():

    annotations = [
        Annotation(start=1, end=10, bitscore=20),
        Annotation(start=5, end=10, bitscore=30),
        Annotation(start=9, end=15, bitscore=10),
        Annotation(start=16, end=20, bitscore=30),
        Annotation(start=16, end=20, bitscore=20),
        Annotation(start=21, end=30, bitscore=40),
    ]
    choose_func = functools.partial(choose_annotation, overlap=2)

    assert filter_annotations(annotations, choose_func=choose_func) == \
        set(
            [
                annotations[1],
                annotations[2],
                annotations[3],
                annotations[5],
            ]
        )


def test_filter_annotations1():

    annotations = [
        Annotation(start=1, end=10, bitscore=20),
        Annotation(start=5, end=10, bitscore=30),
        Annotation(start=9, end=15, bitscore=10),
        Annotation(start=16, end=20, bitscore=30),
        Annotation(start=16, end=20, bitscore=20),
        Annotation(start=21, end=30, bitscore=40),
    ]
    choose_func = None

    assert filter_annotations(annotations, choose_func=choose_func) == \
        set(
            [
                annotations[1],
                annotations[2],
                annotations[3],
                annotations[5],
            ]
        )


def test_filter_base_ok1():
    a = Annotation(seq_id='seq1')
    assert filter_base(a, 'seq_id', 'seq1')


def test_filter_base_fail1():
    a = Annotation(seq_id='seq1')
    assert not filter_base(a, 'seq_id', 'seq')


def test_filter_base_fail2():
    a = Annotation(seq_id='seq1')
    assert not filter_base(a, 'seqid', 'seq1')


def test_filter_len_ge_ok1():
    # equal
    a = Annotation(start=1, end=2)
    assert filter_len(a, 2, True)


def test_filter_len_le_ok1():
    # equal
    a = Annotation(start=1, end=2)
    assert filter_len(a, 2, False)


def test_filter_len_ge_ok2():
    # greater
    a = Annotation(start=1, end=3)
    assert filter_len(a, 2, True)


def test_filter_len_le_ok2():
    # lower
    a = Annotation(start=1, end=3)
    assert filter_len(a, 5, False)


def test_filter_len_ge_fail1():
    a = Annotation(start=1, end=3)
    assert not filter_len(a, 5, True)


def test_filter_len_le_fail1():
    a = Annotation(start=1, end=10)
    assert not filter_len(a, 5, False)


def test_filter_base_num_ge_ok1():
    a = Annotation(bitscore=10)
    assert filter_base_num(a, 'bitscore', 5, True)


def test_filter_base_num_ge_ok2():
    a = Annotation(bitscore=10)
    assert filter_base_num(a, 'bitscore', 10, True)


def test_filter_base_num_ge_fail1():
    a = Annotation(bitscore=10)
    assert not filter_base_num(a, 'bitscore', 15, True)


def test_filter_base_num_ge_fail2():
    a = Annotation()
    assert not filter_base_num(a, 'bitscore', 15, True)


def test_filter_base_num_le_ok1():
    a = Annotation(bitscore=10)
    assert filter_base_num(a, 'bitscore', 15, False)


def test_filter_base_num_le_ok2():
    a = Annotation(bitscore=10)
    assert filter_base_num(a, 'bitscore', 10, False)


def test_filter_base_num_le_fail1():
    a = Annotation(bitscore=10)
    assert not filter_base_num(a, 'bitscore', 5, False)


def test_filter_base_num_le_fail2():
    a = Annotation()
    assert not filter_attr_num(a, 'bitscore', 15, False)


def test_filter_attr_num_ge_ok1():
    a = Annotation(bitscore=10)
    assert filter_attr_num(a, 'bitscore', 5, True)


def test_filter_attr_num_ge_ok2():
    a = Annotation(bitscore=10)
    assert filter_attr_num(a, 'bitscore', 10, True)



def test_filter_attr_num_ge_fail1():
    a = Annotation(bitscore=10)
    assert not filter_attr_num(a, 'bitscore', 15, True)


def test_filter_attr_num_ge_fail2():
    a = Annotation()
    assert not filter_attr_num(a, 'bitscore', 15, True)


def test_filter_attr_num_le_ok1():
    a = Annotation(bitscore=10)
    assert filter_attr_num(a, 'bitscore', 15, False)


def test_filter_attr_num_le_ok2():
    a = Annotation(bitscore=10)
    assert filter_attr_num(a, 'bitscore', 10, False)


def test_filter_attr_num_le_fail1():
    a = Annotation(bitscore=10)
    assert not filter_attr_num(a, 'bitscore', 5, False)


def test_filter_attr_num_le_fail2():
    a = Annotation()
    assert not filter_attr_num(a, 'bitscore', 5, False)


@pytest.mark.parametrize(
    "annotation,attr,value,greater,result",
    [
        (Annotation(bitscore=11), 'bitscore', 10, True, True),
        (Annotation(bitscore=10), 'bitscore', 10, True, False),
        (Annotation(bitscore=None), 'bitscore', 10, True, False),
        (Annotation(bitscore=11), 'bitscore', 10, False, False),
        (Annotation(bitscore=9), 'bitscore', 10, False, True),
    ]
)
def test_filter_attr_num_s(annotation, attr, value, greater, result):
    assert filter_attr_num_s(annotation, attr, value, greater) == result


def test_filter_attr_str_eq_ok1():
    a = Annotation(ec='1.1.1.1')
    assert filter_attr_str(a, 'ec', '1.1.1.1', True)


def test_filter_attr_str_eq_fail1():
    a = Annotation(ec='1.1.1.1')
    assert not filter_attr_str(a, 'ec1', '1.1.1.1', True)


def test_filter_attr_str_eq_fail2():
    a = Annotation(ec='1.1.1.1')
    assert not filter_attr_str(a, 'ec', '1.1.1', True)


def test_filter_attr_str_in_ok1():
    a = Annotation(ec='1.1.1.1')
    assert filter_attr_str(a, 'ec', '1.1.1.1', False)


def test_filter_attr_str_in_ok2():
    a = Annotation(ec='1.1.1.1')
    assert filter_attr_str(a, 'ec', '1.1.1', False)


def test_filter_attr_str_in_fail1():
    a = Annotation(ec='1.1.1.1')
    assert not filter_attr_str(a, 'ec1', '1.1.1.1', False)


def test_filter_attr_str_in_fail2():
    a = Annotation(ec='1.1.1.1')
    assert not filter_attr_str(a, 'ec', '1.1.1.2', False)



"""
.. versionadded:: 0.1.14

Code deprecated or untested
"""

from __future__ import division

import matplotlib.pyplot as plt
import matplotlib.cm as cm
from ..utils.common import deprecated
import numpy
import logging

try:
    from scipy.spatial import ConvexHull
except ImportError:
    ConvexHull = None

try:
    import seaborn as sns
except ImportError:
    sns = None

LOG = logging.getLogger(__name__)

TAXON_COLOURS = {
    'archaea': '#4daf4a',
    'bacteria': '#377eb8',
    'fungi': '#ff7f00',
    'streptophyta': '#e41a1c',
    'ciliophora': '#984ea3'
}
"Default colours for root taxa"

# for backward compatibility
TAXON_COLORS = TAXON_COLOURS


@deprecated
def lineplot_values_on_second_axis(gene_num, axis, colour='c', ylabel=''):
    """
    .. deprecated:: 0.1.13

    Adds a lineplot on a second axis using twinx
    """
    ax2 = axis.twinx()
    ax2.plot(axis.get_xticks(), gene_num, colour)

    ax2.set_ylabel(ylabel)
    ax2.set_yscale('log')

    return ax2


def plot_scatter_3d(data, labels, colours=None, pointsize=10., title='',
                    xlabel='', ylabel='', zlabel=''):
    """
    Scatter plot in 3d. Used for cluster results

    :param array data: :class:`numpy.array` with shape n, 3
    :param array labels: labels to categorise samples
    :param dict colours: dictionary whose keys are the labels and the values
        are valid :mod:`matplotlib` colours
    :param int pointsize: point size
    :param str title: plot title
    :param str xlabel: label for x axis
    :param str ylabel: label for y axis
    :param str zlabel: label for z axis

    :return: axis instance
    """

    fig = plt.figure()
    ax = plt.subplot(111, projection='3d')
    label_dict = {}
    for label, features in zip(labels, data):
        try:
            label_dict[label].append(features)
        except KeyError:
            label_dict[label] = [features]

    for label, data_points in label_dict.iteritems():
        if colours is None:
            colour = numpy.random.rand(3, 1)
        else:
            colour = colours[label]
        data_points = numpy.array(data_points)
        ax.scatter(data_points[:, 0], data_points[:, 1], data_points[:, 2],
                   s=pointsize, c=colour)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_zlabel(zlabel)
    fig.suptitle(title)

    return ax


def plot_scatter_2d(data, labels, colours=None, pointsize=10., title='',
                    xlabel='', ylabel='', centers=None, marker='*',
                    marker_colour=None, markersize=None, hull_points=True,
                    linewidth=0.2, legend=True, anno_center=True):
    """
    Scatter plot in 2d. Used for cluster results

    :param array data: :class:`numpy.array` with shape n, 2
    :param array labels: labels to categorise samples
    :param dict colours: dictionary whose keys are the labels and the values
        are valid :mod:`matplotlib` colours
    :param int pointsize: point size
    :param str title: plot title
    :param str xlabel: label for x axis
    :param str ylabel: label for y axis

    :return: axis instance
    """
    fig = plt.figure()
    ax = plt.subplot(111)
    label_dict = {}
    for label, features in zip(labels, data):
        try:
            label_dict[label].append(features)
        except KeyError:
            label_dict[label] = [features]

    if colours is None:
        # unify this, generate random colors in a go
        gen_colours = []

    colour_step = 256 // len(label_dict)

    for label, data_points in sorted(label_dict.iteritems()):
        if colours is None:
            colour = cm.Set1(label * colour_step)
            gen_colours.append(colour)
        else:
            colour = colours[label]
        data_points = numpy.array(data_points)
        if (len(data_points) > 2) and hull_points:
            if ConvexHull is None:
                LOG.error("ConvexHull from scipy.spatial couldn't be imported")
            else:
                hull = ConvexHull(data_points)
                for simplex in hull.simplices:
                    ax.plot(
                        data_points[simplex, 0],
                        data_points[simplex, 1],
                        '-',
                        colour=colour
                    )
        ax.scatter(
            data_points[:, 0],
            data_points[:, 1],
            s=pointsize,
            c=colour,
            linewidth=linewidth
        )

    if centers is not None:
        for idx, (center_x, center_y) in enumerate(centers):
            ax.plot(
                center_x,
                center_y,
                marker,
                markerfacecolor=marker_colour if marker_colour else gen_colours[idx],
                markeredgecolor='k',
                markersize=pointsize*1.2 if markersize is None else markersize,
                label=idx
            )
            if anno_center:
                ax.annotate(idx, xy=(center_x, center_y))

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    fig.suptitle(title)

    if legend:
        ax.legend()

    return ax


def barchart_categories(data, colours=None, title='', tickfont='small',
                        xlabel_dict=None, barlabel_dict=None, width=0.9,
                        rotation='vertical', file_name=None, fig_size=None,
                        fig_aspect=None):
    """
    :param data: DataFrame where the number of rows indicates how many bars
        will plotted per column
    :param colours: must be equal the number of data rows if supplied or it
        will be blue by default
    :param title: chart title
    :param tickfont: font size for ticks (only for column axis)
    :param xlabel_dict: a mapping to the acual labels to use for the columns.
        Defaults to columns' names
    :param barlabel_dict: a mapping to the acual labels to use for the row.
        Defaults to rows' names
    :param width: bar width

    :return: axis instance
    """

    if colours is None:
        colours = ('b',) * len(data.index)

    if fig_aspect is not None:
        fig_size = plt.figaspect(fig_aspect)
    fig = plt.figure(dpi=300, figsize=fig_size)

    ax = fig.add_subplot(111)

    base_index = numpy.arange(0.,
                              len(data.columns) * len(data.index),
                              len(data.index))

    for idx, (row_name, series) in enumerate(data.iterrows()):
        ax.bar(
            base_index + idx * width,
            series,
            color=colours[idx],
            label=row_name if barlabel_dict is None else barlabel_dict[idx]
        )

    ax.set_xticks(base_index + len(data.index) / 2.)

    xlabels = ax.set_xticklabels(
        [
            col_id if xlabel_dict is None else xlabel_dict[col_id]
            for col_id in data.columns
        ],
        rotation=rotation
    )

    ax.legend()

    ax.set_xlim(0, max(base_index) + (width * len(data.index)))
    ax.set_title(title)

    if file_name is not None:
        fig.savefig(file_name, bbox_inches='tight',
                    bbox_extra_artist=(xlabels,))
        fig.clf()

    return ax


def plot_contig_assignment_bar(series, taxon_colours=None, log_scale=False,
                               index=None, file_name=None, fig_aspect=None,
                               xlabels_size=8):
    """
    Plots barchart for contig assignment

    :param series: :class:`pandas.Series` instance with the data
    :param dict taxon_colours: colour of the bars for each taxon
    :param bool log_scale: if True the y axis is log scaled
    :param tuple fig_aspect: tuple with figure size
    :param int xlabels_size: size of the taxon labels
    :param index: optional :class:`pandas.Index` used to reindex the series
    :param str file_name: name of the file to write the graph to
    """
    LOG.info(
        "Plotting %d taxa, using colours (%s) and log scale (%s)",
        len(series),
        'no' if not taxon_colours else 'yes',
        'no' if not log_scale else 'yes'
    )

    if index is not None:
        series = series.reindex(index)

    if fig_aspect is None:
        fig_aspect = plt.figaspect(2. / len(series))

    LOG.info(
        "Graph size %r and label size %d",
        tuple(fig_aspect),
        xlabels_size
    )

    fig = plt.figure(dpi=300, figsize=fig_aspect)

    ax = plt.subplot(111)

    colours = [taxon_colours[taxon_name] for taxon_name in series.index]

    ax.bar(range(1, series.count() + 1), series, color=colours,
           log=log_scale, align='center')

    ax.set_xlim(0, series.count() + 1)

    ax.set_xticks(range(1, series.count() + 1))
    xlabels = ax.set_xticklabels(series.index, rotation='vertical',
                                 fontsize=xlabels_size)
    ax.grid(True, axis='y')
    ax.set_ylabel('Number of contigs assigned', fontsize='medium',
                  horizontalalignment='right')
    ax.set_xlabel('Taxa', fontsize='medium')

    if file_name is not None:
        LOG.info("Saving graph to file %s", file_name)
        fig.savefig(file_name, bbox_inches='tight',
                    bbox_extra_artist=(xlabels,))
        fig.clf()

    return ax


def scatter_gene_values(gene_dict, xlabel="Profile pN/pS",
                        ylabel="Rumen pN/pS", title="", colours=None,
                        file_name=None, plot_order=None, line_colour='r',
                        max_limit=None, axes=None):
    """
    Plots gene-taxon pN/pS from profiles against their observed values.

    :param dict gene_dict: dictionary that contains the data
    :param str xlabel: label for x axis
    :param str ylabel: label for y axis
    :param str title: graph title
    :param colours: colours used in for the different datasets; defaults to
        TAXON_COLOURS
    :param str file_name: path to which the graph is to be saved (by default)
        it doesn't write to disk
    :param iterable plot_order: the order used in plotting the data points;
        default to the order of the gene_dict dictionary keys
    :param coloUr: valid colour for the lines in the plot
    :param float max_limit: used to put a limit on the plot
    :param axes: optional axes used to draw the scatter plot

    :return: the axis object used for the plot
    """
    if colours is None:
        colours = TAXON_COLOURS

    if plot_order is None:
        plot_order = gene_dict.keys()

    if axes is None:
        fig = plt.figure()
        ax = plt.subplot(111)
    else:
        ax = axes

    for root in plot_order:
        root_values = gene_dict[root]
        values = numpy.array(root_values)

        if values.size == 0:
            continue

        ax.scatter(values[:, 0], values[:, 1], color=colours[root], label=root)

    if max_limit is None:
        max_limit = max(ax.get_xlim()[1], ax.get_ylim()[1])
    ax.plot(numpy.arange(0., max_limit + 1), numpy.arange(0., max_limit + 1),
            color=line_colour)

    ax.axvline(1.0, color=line_colour)
    ax.axhline(1.0, color=line_colour)

    ax.set_xlim(left=0., right=max_limit)
    ax.set_ylim(bottom=0., top=max_limit)

    ax.legend()
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)

    if file_name is not None:
        LOG.info("Saving graph to file %s", file_name)
        fig.savefig(file_name)
        fig.clf()

    return ax


def map_taxon_to_colours(taxa, taxonomy, default_colour='#ffff33'):
    """
    Returns a dictionary of taxa and their assigned colours based on
    :data:`TAXON_COLOURS` and the taxonomy provided. Uses the
    :func:`taxon.Taxonomy.get_taxon_root` to determine the root of a
    taxon.

    :param iterable taxa: iterable of taxon ids
    :param Taxonomy taxonomy: taxonomy instance
    :param default_colour: colour used in case there's no known root for the
        taxon

    :return dict: dictionary mapping taxon_id to colour
    """
    LOG.info("Assigning taxa colours")
    tx_colours = {}

    for taxon_id in taxa:
        try:
            tx_colours[taxon_id] = TAXON_COLOURS[
                taxonomy.get_taxon_root(taxon_id).s_name
            ]
        except KeyError:
            tx_colours[taxon_id] = default_colour

    return tx_colours


get_taxon_colors_new = map_taxon_to_colours



"""
.. versionadded:: 0.1.15

Module to plot relative abundances in a 1D or 3D projection
"""
from __future__ import division
from builtins import zip
import math

import numpy

SQRT3_2 = math.sqrt(3) / 2


def draw_triangle_grid(ax, labels=['LAM', 'SAM', 'EAM'], linewidth=1.,
                       styles=['-', ':', '--'], fontsize=22):
    """
    .. versionchanged:: 0.2.0
        reworked internals and changed defaults

    Draws a triangle as axes, for a planar-simplex projection.

    Arguments:
        ax: an axis instance
        labels (iterable): list of string to be put for the axes
        styles (None, iterable): either *None* for solid lines or matplotlib
            line markers. These are in sync between the internal lines and
            the axes.
        linewidth (float): line width for the axes, the internal lines are
            equal to *0.75 * linewidth*
        fontsize (float): font size for the labels, the tick font size is
            equal to *0.75 * fontsize*
    """

    ax.set_ylim((-0.1, 1.0))
    ax.set_xlim((-0.1, 1.1))
    ax.set_axis_off()

    # SAM
    ax.text(0, -0.05, labels[0], fontsize=fontsize, ha='right')
    # LAM
    ax.text(1.0, -0.05, labels[1], fontsize=fontsize, ha='left')
    # EAB
    ax.text(0.5, SQRT3_2 + 0.025, labels[2], fontsize=fontsize, ha='center')

    values = numpy.arange(0.1, 1., 0.1)

    points = []

    # Bottom axis (LAM)
    x1, y1 = project_point((1, 0, 0))
    x2, y2 = project_point((0, 1, 0))
    points.append((
        [x1, x2],
        [y1, y2]
    ))
    # Left axis (EAM)
    x1, y1 = project_point((1, 0, 0))
    x2, y2 = project_point((0, 0, 1))
    points.append((
        [x1, x2],
        [y1, y2]
    ))
    # Right axis (SAM)
    x1, y1 = project_point((0, 1, 0))
    x2, y2 = project_point((0, 0, 1))
    points.append((
        [x1, x2],
        [y1, y2]
    ))

    if styles is None:
        styles = ['-'] * 3

    for (X, Y), style in zip(points, styles):
        ax.plot(X, Y, linestyle=style, linewidth=linewidth*1.5, color='k')

    for index in values:

        x1, y1 = project_point((index, 1 - index, 0))
        x2, y2 = project_point((0, 1 - index, index))
        # EAM
        ax.plot([x1, x2], [y1, y2], linestyle=styles[2], color='k',
                linewidth=linewidth)

        x3, y3 = project_point((0, index, 1 - index))
        x4, y4 = project_point((index, 0, 1 - index))
        # SAM
        ax.plot([x3, x4], [y3, y4], linestyle=styles[1], color='k',
                linewidth=linewidth)

        # LAM
        ax.plot([x4, x1], [y4, y1], linestyle=styles[0], color='k',
                linewidth=linewidth)

        # LAM labels
        ax.text(x1, y1 - 0.02, numpy.round(index, 1), va='top', ha='center',
                fontsize=int(fontsize * 0.75))
        # EAM labels
        ax.text(x4 - 0.02, y4, numpy.round(1 - index, 1), ha='right', va='center',
                fontsize=int(fontsize * 0.75))
        # EAM labels
        ax.text(x3 + 0.02, y3, numpy.round(index, 1), ha='left', va='center',
                fontsize=int(fontsize * 0.75))


def draw_1d_grid(ax, labels=['LAM', 'SAM'], fontsize=22):
    """
    .. versionchanged:: 0.2.0
        reworked internals and changed defaults

    Draws a 1D axis, to display propotions.

    Arguments:
        ax: an axis instance
        labels (iterable): list of string to be put for the axes
        fontsize (float): font size for the labels, the tick font size is
            equal to *0.75 * fontsize*
    """

    ax.set_ylim((-1.1, 1.1))
    ax.set_xlim((-0.1, 1.1))
    ax.set_axis_off()

    ax.text(-0.025, 0, labels[0], fontsize=fontsize, va='center', ha='right')
    ax.text(1.025, 0, labels[1], fontsize=fontsize, va='center', ha='left')

    ty = 1.0
    by = -1.0

    ax.plot([0, 1], [0, 0], linewidth=1, color='k')

    for x in numpy.arange(0.1, 1., 0.1):
        ax.plot([x, x], [by, ty], color='k', linestyle=':')
        y_b = by - .1
        y_t = ty + .1
        ax.text(x, y_b, numpy.round(x, 1), fontsize=fontsize * 0.75, ha='center', va='top')
        ax.text(x, y_t, numpy.round(1 - x, 1), fontsize=fontsize * 0.75, ha='center',
                va='bottom')


def project_point(point):
    """
    Project a tuple containing coordinates (i.e. x, y, z) to planar-simplex.

    Arguments:
        point (tuple): contains the three coordinates to project

    Returns:
        tuple: the projected point in a planar-simplex
    """
    # a = point[0]
    b = point[1]
    c = point[2]

    x = b + (c / 2.)
    y = SQRT3_2 * c

    return (x, y)


def col_func_firstel(key, colors=None):
    if colors is None:
        return 'black'

    return colors[key[0]]


def col_func_name(key, func=None, colors=None):
    if (colors is None):
        return 'black'

    if func is not None:
        key = func(key)

    return colors[key]


def col_func_taxon(taxon_id, taxonomy, anc_ids, colpal):
    for anc_id, color in zip(anc_ids, colpal):
        if taxonomy.is_ancestor(taxon_id, anc_id):
            return color
    return 'black'


def draw_circles(ax, data, col_func=col_func_name, csize=200, alpha=.5,
                 sizescale=None, order=None, linewidths=0., edgecolor='none'):
    """
    .. versionchanged:: 0.2.0
        changed internals and added return value

    Draws a scatter plot over either a planar-simplex projection, if the number
    of coordinates is 3, or in a 1D axis.

    If the number of coordinates is 3, :func:`project_point` is used to project
    the point in 2 coordinates. The coordinates are converted in proportions
    internally.

    Arguments:
        ax: axis to plot on
        data (pandas.DataFrame): a DataFrame with 2 for a 1D plot or 3 columns
            for a planar-simplex
        col_func (func): a function that accept a parameter, an element of the
            DataFrame index and returns a colour for it
        csize (int): the base size of the circles
        alpha (float): transparency of the circles, between 0 and 1 included
        sizescale (None, pandas.Series): a Series or dictionary with the same
            elements as the Index of *data*, whose values are the size factors
            that are multiplied to *csize*. If **None**, the size of the
            circles is equal to *csize*
        order (None, iterable): iterable with the elements of *data* Index, to
            specify the order in which the circles must be plotted. If None,
            the order is the same as *data.index*
        linewidths (float): width of the circle line
        edgecolor (str): color of the circle line

    Returns:
        PathCollection: the return value of matplotlib *scatter*

    .. note::

        To **not** have circle lines, *edgecolor* must be *'none'* and
        *linewidths* equal *0*

    """
    # normalize data (ternary plots requires that the sum of a row is 1)
    data = data.div(data.sum(axis=1), axis=0)

    def no_project(a, b):
        if a > b:
            x = 1 - a
        else:
            x = b
        return x, 0.

    if order is None:
        order = data.index

    X = []
    Y = []
    for key, coord in data.loc[order].iterrows():
        # print coord
        x, y = project_point(coord) if len(coord) == 3 else no_project(*coord)
        X.append(x)
        Y.append(y)

    colors = [col_func(key) for key in order]

    paths = ax.scatter(
        X,
        Y,
        color=colors,
        alpha=alpha,
        linewidths=linewidths,
        edgecolor=edgecolor,
        marker=u'o',
        s=csize * (sizescale.loc[order] if sizescale is not None else 1)
    )
    return paths


def draw_axis_internal_triangle(ax, color='r', linewidth=2.):
    """
    .. versionadded:: 0.2.5

    Draws a triangle that indicates the 50% limit for all 3 samples

    Arguments:
        ax: axis to use
        color (str, float, tuple): color used to draw the triangle
        linewidth (float): line width
    """
    x1, y1 = project_point([.5, .5, 0.])
    x2, y2 = project_point([.5, 0., 0.5])
    x3, y3 = project_point([.0, 0.5, 0.5])

    ax.plot(
        [x1, x2, x3, x1],
        [y1, y2, y3, y1],
        color=color,
        linewidth=linewidth
    )



"""
Mapping functions for SNPs - Should be move into an 'iterator' package to
be shared with other modules?
"""


def map_gene_id(gene_id, gene_map=None):
    """
    Returns an iterator for all the values of a dictionary. if gene_id is not
    found in the gene_map, an empty iterator is returned.

    Arguments:
        gene_id (immutable): gene_id or any other dictionary key.
        gene_map (dict): a dictionary in the form key->[v1, v2, .. vN]

    Returns:
        generator: iterator (empty if gene_id is not in gene_map) with the
        values

    """
    if gene_id not in gene_map:
        return
    for map_id in gene_map[gene_id]:
        yield map_id


def map_taxon_id_to_rank(taxon_id, rank=None, taxonomy=None,
                         include_higher=False):
    """
    Given a taxon_id, returns an iterator with only the element that correspond
    to the requested rank. If the taxon returned by
    :class:`mgkit.taxon.Taxonomy.get_ranked_taxon` has a different rank
    than requested, the iterator will be empty if `include_higher` is False
    and the returned taxon ID if True.

    Arguments:
        taxon_id (int): taxon ID to be mapped
        rank (str): taxon rank used (:data:`mgkit.taxon.TAXON_RANKS`)
        include_higher (bool): determines if a rank higher than the one
            requested is to be returned

    Returns:
        generator: iterator with the values or empty
    """
    ranked_taxon = taxonomy.get_ranked_taxon(taxon_id, rank=rank)

    # in case we don't want to include higher taxa
    if (ranked_taxon.rank != rank) and (not include_higher):
        return

    yield ranked_taxon.taxon_id


def map_taxon_id_to_ancestor(taxon_id, anc_ids=None, func=None):
    """
    Given a taxon_id and a list of ancestors IDs, returns an iterator with the
    IDs that are ancestors of taxon_id.

    Arguments:
        taxon_id (int): taxon ID to be mapped
        anc_ids (iterable): taxon IDs to check for ancestry
        func: function used to check for ancestry - partial function for
            :func:`mgkit.taxon.is_ancestor` that accepts taxon_id and anc_id

    Returns:
        generator: iterator with the values or empty

    .. note::

        check :func:`mgkit.filter.taxon.filter_taxon_by_id_list` for examples
        on using func

    """
    for anc_id in anc_ids:
        if func(taxon_id, anc_id):
            yield anc_id



"""
Blast output conversion in GFF requires a BLAST+ tabular format which can be
obtained by using the `--outfmt 6` option with the default columns, as
specified in :func:`mgkit.io.blast.parse_blast_tab`. The script can get data
from the standard in and ouputs GFF lines on the standard output by default.

Uniprot
*******

The Function :func:`mgkit.io.blast.parse_uniprot_blast` is used, which filters
BLAST hits based on bitscore and adds by default a *db* attribute to the
annotation with the value *UNIPROT-SP*, indicating that the SwissProt db is
used and a *dbq* attribute with the value 10. The feature type used in the GFF
is CDS.

.. blockdiag::

    {
        "BLAST+" [color = "#377eb8" , textcolor = 'white',
            shape = flowchart.input];
        "parse_uniprot_blast" [color = "#e41a1c" , textcolor = 'white',
            width=200, fontsize=16];
        "GFF" [color = "#4daf4a" , textcolor = 'white'];
        "BLAST+"  -> "parse_uniprot_blast" -> GFF;
    }

BlastDB
*******

If a BlastDB, such as *nt* or *nr* was used, the **blastdb** command offers
some quick defaults to parse BLAST results.

It now includes options to control the way the sequence header are formatted.
Options to change the separator used, as well as the column used as *gene_id*.
This was added because at the moment the GI identifier (the second column in
the header) is used, but it's being phased out in favour of the embl/gb/dbj
(right now the fourth column in the header). This should easy the transition to
the new format and makes it easier to adapt an older pipeline/blastdb to newer
files (like the ID to TAXA files).

The header from the a *ncbi-nt* header looks like this::

    gi|160361034|gb|CP000884.1

This is the default output accepted by the *blastdb* command. The fields are
separated by **|** (pipe) and the GI is used (--gene-index 1, since internally
the string is split by the separator and the second element is take - lists
indices are 0-based in Python). This output uses the following options::

    --header-sep '|' --gene-index 1

Notice the single quotes to pass the pipe symbol, since *bash* would interpret
it as pipeing to the next coommand otherwise. This is the default.

In case, for the same header, we want to use the *gb* identifier, the only
option to be specified is::

    --gene-index 3

This will get the fourth element of the header (since we're splitting by pipe).

As in the *uniprot* command, the *gene_id* can be set to use the whole header,
using the *-n* option. Useful in case the *BLAST* db that was used was custom
made. While pipe is used in major databases, it was made the default, by if the
db used has different conventions the separator can be changed. There's also
the options of later changing the *gene_id* in the output GFF if necessary.

Changes
*******

.. versionchanged:: 0.3.4
    using *click* instead of *argparse*

.. versionchanged:: 0.2.6
    added *-r* option to *blastdb*

.. versionchanged:: 0.2.5
    added more options to give user control to the *blastdb* command

.. versionadded:: 0.2.3
    added *--fasta-file* option, added more data from a blsat hit

.. versionadded:: 0.2.2
    added *blastdb* command

.. versionchanged:: 0.2.1

    added *-ft* option

.. versionchanged:: 0.1.13
    added *-n* and *-k* parameters to *uniprot* command

.. versionadded:: 0.1.12

"""

import logging
import click
from tqdm import tqdm
from .. import logger
from . import utils
from ..io import blast, fasta


LOG = logging.getLogger(__name__)


@click.group()
@click.version_option()
@utils.cite_option
def main():
    "Main function"
    pass


def load_fasta_file(file_name):
    if file_name is None:
        return None

    return dict(
        (name, len(seq))
        for name, seq in fasta.load_fasta_rename(file_name)
    )


def validate_params(ctx, param, values):
    new_values = []
    for value in values:
        value = value.split(':')
        if len(value) != 2:
            raise click.BadParameter(
                "Wrong key/value format, must be 'key:value' 'key:value'"
            )
        new_values.append(value)
    return new_values


@main.command('blastdb', help="""
Reads a BLAST output file [blast-file] in tabular format (using -outfmt 6) and
outputs a GFF file [gff-file]
""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-db', '--db-used', default='NCBI-NT', show_default=True,
              help='blastdb used')
@click.option('-n', '--no-split', is_flag=True,
              help='''if used, the script assumes that the sequence header will be used as gene_id''')
@click.option('-s', '--header-sep', default='|', show_default=True,
              help="""The separator for the header, defaults to '|' (pipe)""")
@click.option('-i', '--gene-index', default=1, type=click.INT, show_default=True,
              help="""Which of the header columns (0-based) to use as gene_id (defaults to 1 - the second column)""")
@click.option('-r', '--remove-version', is_flag=True,
              help='''if used, the script removes the *version* information from the gene_id''')
@click.option('-a', '--fasta-file', type=click.Path(readable=True),
              help='Optional FASTA file with the query sequences')
@click.option('-dbq', '--db-quality', type=click.INT, show_default=True,
              default=10, help='Quality of the DB used')
@click.option('-b', '--bitscore', type=click.FLOAT, show_default=True,
              help='Minimum bitscore to keep the annotation', default=0.0)
@click.option('-k', '--attr-value', multiple=True, callback=validate_params,
              default=None, help='''Additional attribute and value to add to each annotation, in the form attr:value''')
@click.option('-ft', '--feat-type', show_default=True, default='CDS',
              help='Feature type to use in the GFF')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('blast-file', type=click.File('rb', lazy=False), default='-')
@click.argument('gff-file', type=click.File('wb', lazy=False), default='-')
def convert_from_blastdb(verbose, db_used, no_split, header_sep, gene_index,
                         remove_version, fasta_file, db_quality, bitscore,
                         attr_value, feat_type, progress, blast_file,
                         gff_file):
    """
    .. versionadded:: 0.2.2
    """

    logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(gff_file, 'name', repr(gff_file))
    )

    seqs = load_fasta_file(fasta_file)

    if no_split:
        def name_func(x): return x
    else:
        if remove_version:
            def name_func(x):
                return x.split(header_sep)[gene_index].split('.')[0]
        else:
            def name_func(x):
                return x.split(header_sep)[gene_index]

    iterator = blast.parse_uniprot_blast(
        blast_file,
        bitscore=bitscore,
        db=db_used,
        dbq=db_quality,
        name_func=name_func,
        feat_type=feat_type,
        seq_lengths=seqs
    )

    if progress:
        iterator = tqdm(iterator)

    for annotation in iterator:
        if attr_value is not None:
            for key, value in attr_value:
                annotation.set_attr(key, value)

        annotation.to_file(gff_file)


@main.command('uniprot', help="""
Reads a BLAST output file [blast-file] in tabular format (using -outfmt 6) from
a Uniprot DB and outputs a GFF file [gff-file]
""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-db', '--db-used', default='UNIPROT-SP', show_default=True,
              help='Uniprot database used with BLAST')
@click.option('-n', '--no-split', is_flag=True,
              help='''if used, the script assumes that the sequence header will be used as gene_id''')
@click.option('-a', '--fasta-file', type=click.Path(readable=True),
              help='Optional FASTA file with the query sequences')
@click.option('-dbq', '--db-quality', type=click.INT, show_default=True,
              default=10, help='Quality of the DB used')
@click.option('-b', '--bitscore', type=click.FLOAT, show_default=True,
              default=0.0, help='Minimum bitscore to keep the annotation')
@click.option('-k', '--attr-value', multiple=True, callback=validate_params,
              default=None, help='''Additional attribute and value to add to each annotation, in the form attr:value''')
@click.option('-ft', '--feat-type', default='CDS', show_default=True,
              help='Feature type to use in the GFF')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('blast-file', type=click.File('rb', lazy=False), default='-')
@click.argument('gff-file', type=click.File('wb', lazy=False), default='-')
def convert_from_uniprot(verbose, db_used, no_split, fasta_file, db_quality,
                         bitscore, attr_value, feat_type, progress, blast_file,
                         gff_file):

    logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(gff_file, 'name', repr(gff_file))
    )

    seqs = load_fasta_file(fasta_file)

    if no_split is True:
        def name_func(x): return x
    else:
        name_func = None

    iterator = blast.parse_uniprot_blast(
        blast_file,
        bitscore=bitscore,
        db=db_used,
        dbq=db_quality,
        name_func=name_func,
        feat_type=feat_type,
        seq_lengths=seqs
    )

    if progress:
        iterator = tqdm(iterator)

    for annotation in iterator:
        if attr_value is not None:
            for key, value in attr_value:
                annotation.set_attr(key, value)

        annotation.to_file(gff_file)



"""
.. versionadded:: 0.1.9

Taxa filtering functions
"""

from .common import FilterFails
from ..taxon import is_ancestor
import functools


def filter_taxon_by_id_list(taxon_id, filter_list=None, exclude=False,
                            func=None):
    """
    Filter a taxon_id against a list of taxon ids. Returns True if the
    conditions of the filter are met.

    If func is not None, a function that accepts two values is expected,
    it should be either a partial `is_ancestor` which only accepts taxon_id and
    anc_id or another function that behaves the same way.

    .. note::

            if func is None, a simple lambda is used to test identity::

                func = lambda t_id, a_id: t_id == a_id

    Arguments:
        taxon_id (int): the taxon id to filter
        filter_list (iterable): an iterable with taxon ids
        exclude (bool): if the filter is reversed (i.e. included if NOT found)
        func (func or None): a function that accepts taxon_id and an anc_id
            and returns a bool to indicated if anc_id is ancestor of taxon_id.
            Equivalent to :func:`~mgkit.taxon.is_ancestor`.

    Returns:
        bool:
            True if the taxon_id is in the filter list (or a descendant of it)
            False if it's not found. Exclude equal to True reverse the result.

            +-------+---------+--------------+
            | Found | Exclude | Return Value |
            +=======+=========+==============+
            | Yes   | False   | True         |
            +-------+---------+--------------+
            | No    | False   | False        |
            +-------+---------+--------------+
            | Yes   | True    | False        |
            +-------+---------+--------------+
            | No    | True    | True         |
            +-------+---------+--------------+

    Example:

        If using func and assuming that `taxonomy` is an instance of
        :class:`~mgkit.taxon.Taxonomy` with data loaded:

        >>> import functools
        >>> import mgkit.taxon
        >>> func = functools.partial(mgkit.taxon.is_ancestor, taxonomy)
        >>> filter_taxon_by_id_list(1200582, [838], func=func)
        True

    """

    if func is None:
        def func(t_id, a_id): return t_id == a_id

    if filter_list is None:
        raise FilterFails('No filter_list')

    ret_val = any(
        func(taxon_id, filter_id)
        for filter_id in filter_list
    )

    return ret_val ^ exclude


def filter_by_ancestor(taxon_id, filter_list=None, exclude=False,
                       taxonomy=None):
    """
    .. versionadded:: 0.1.13

    Convenience function for :func:`filter_taxon_by_id_list`, as explained in
    the latter example.
    """
    if filter_list is None:
        raise FilterFails('No filter_list')

    func = functools.partial(is_ancestor, taxonomy)

    return filter_taxon_by_id_list(
        taxon_id,
        filter_list=filter_list,
        func=func,
        exclude=exclude
    )



"""
Module containing classes and functions to access Kegg data
"""
from builtins import object
from future.utils import viewitems
import sys
import logging
import pickle
import random
import re
from .utils import dictionary as dict_utils
from .net import url_read
from .io import open_file


LOG = logging.getLogger(__name__)
KEGG_REST_URL = 'http://rest.kegg.jp/'


class KeggClientRest(object):
    """
    .. versionchanged:: 0.3.1
        added a *cache* attribute for some methods

    Kegg REST client

    The class includes methods and data to use the REST API provided by Kegg.
    At the moment it provides methods to for 'link', 'list' and 'get'
    operations,

    `Kegg REST API <http://www.kegg.jp/kegg/rest/keggapi.html>`_

    """
    contact = None
    api_url = KEGG_REST_URL
    cpd_re = re.compile(
        r"ENTRY\s+(C\d{5})\s+Compound\nNAME\s+([,.\w+ ()-]+);?"
    )
    rn_name_re = re.compile(r"R\d{5}")
    rn_eq_re = re.compile(r'C\d{5}')
    ko_desc_re = re.compile(
        r"ko:(K\d{5})\t.+?;\s+([\w+, ()/:'\[\]-]+)( \[EC:)?\n?"
    )
    cpd_desc_re = re.compile(
        r"cpd:(C\d{5})\t([\w+, ()\[\]'*.-]+);?\n?"
    )
    id_prefix = {'C': 'cpd', 'k': 'map', 'K': 'ko', 'R': 'rn', 'm': 'path'}

    cache = None

    def __init__(self, cache=None):
        """
        .. versionadded:: 0.3.1

        The "cache" parameter is a file name for the cached data wrote using
        :meth:`KeggClientRest.write_cache`.
        """

        if cache is None:
            self.empty_cache()
        else:
            self.load_cache(cache)

    def empty_cache(self, methods=None):
        """
        .. versionadded:: 0.3.1

        Empties the cache completely or for a specific method(s)

        Arguments:
            methods (iterable, str): string or iterable of strings that are
                part of the cache. If None the cache is fully emptied
        """

        if methods is None:
            methods = ('link_ids', 'get_entry', 'get_ids_names')
        else:
            if isinstance(methods, str):
                methods = [methods]

        if self.cache is None:
            self.cache = {}

        for method in methods:
            self.cache[method] = {}

    def load_cache(self, file_handle):
        """
        .. versionadded:: 0.3.1

        Loads the cache from file
        """
        self.cache = pickle.load(open_file(file_handle, 'rb'))

    def write_cache(self, file_handle):
        """
        .. versionadded:: 0.3.1

        Write the cache to file
        """
        pickle.dump(self.cache, open_file(file_handle, 'wb'))

    # Kegg primitives #

    def link_ids(self, target, kegg_ids, max_len=50):
        """
        .. versionchanged:: 0.3.1
            removed *strip* and cached the results

        The method abstract the use of the 'link' operation in the Kegg API

        The target parameter can be one of the following::

            pathway | brite | module | disease | drug | environ | ko | genome |
            <org> | compound | glycan | reaction | rpair | rclass | enzyme

            <org> = KEGG organism code or T number

        :param str target: the target db
        :param ids: can be either a single id as a string or a list of ids
        :param bool strip: if the prefix (e.g. ko:K00601) should be stripped
        :param int max_len: the maximum number of ids to retrieve with each
            request, should not exceed 50
        :return dict: dictionary mapping requested id to target id(s)
        """
        if isinstance(kegg_ids, str):
            kegg_ids = [kegg_ids]

        if isinstance(kegg_ids, (list, dict)):
            kegg_ids = set(kegg_ids)

        try:
            mapping = self.cache['link_ids'][target]
            LOG.debug('Cached Call')
        except KeyError:
            LOG.debug('Empty Cache')
            mapping = {}
            self.cache['link_ids'][target] = mapping

        download_ids = kegg_ids - set(mapping)
        LOG.debug(
            'Number of cached items (%d/%d)',
            len(kegg_ids) - len(download_ids),
            len(kegg_ids)
        )

        while download_ids:
            try:
                sample_ids = set(random.sample(download_ids, max_len))
            except ValueError:
                # download_ids size is less than max_len
                sample_ids = download_ids.copy()
            if len(kegg_ids) > max_len:
                LOG.info(
                    "Downloading links - %d out of %d",
                    len(kegg_ids) - len(download_ids), len(kegg_ids)
                )
            url = "{0}link/{1}/{2}".format(
                self.api_url, target, '+'.join(sample_ids)
            )
            data = url_read(url, agent=self.contact)
            data = data.split('\n')

            for line in data:
                if not line:
                    continue
                source, linked = line.split('\t')
                source = source.split(':')[1]
                linked = linked.split(':')[1]

                try:
                    mapping[source].append(linked)
                except KeyError:
                    mapping[source] = [linked]

            download_ids = download_ids - sample_ids

        # Empty (Not Found) elements, set to None
        for kegg_id in kegg_ids - set(mapping):
            mapping[kegg_id] = None

        return {
            kegg_id: list(value)
            for kegg_id, value in viewitems(mapping)
            if (kegg_id in kegg_ids) and (value is not None)
        }

    def list_ids(self, k_id):
        """
        The method abstract the use of the 'list' operation in the Kegg API

        The k_id parameter can be one of the following::

            pathway | brite | module | disease | drug | environ | ko | genome |
            <org> | compound | glycan | reaction | rpair | rclass | enzyme

            <org> = KEGG organism code or T number

        :param str k_id: kegg database to get list of ids
        :return list: list of ids in the specified database
        """
        url = "{0}list/{1}".format(self.api_url, k_id)
        data = url_read(url, agent=self.contact)
        # leave out the last \n
        return data[:-1]

    def get_entry(self, k_id, option=None):
        """
        .. versionchanged:: 0.3.1
            this is now cached

        The method abstract the use of the 'get' operation in the Kegg API

        :param str k_id: kegg id of the resource to get
        :param str option: optional, to specify a format
        """
        try:
            data = self.cache['get_entry'][(k_id, option)]
        except KeyError:
            url = "{0}get/{1}/{2}".format(
                self.api_url,
                k_id,
                '' if option is None else option
            )
            data = url_read(url, agent=self.contact)
            self.cache['get_entry'][(k_id, option)] = data

        return data

    def link(self, target, source, options=None):
        """
        .. versionadded:: 0.2.0

        Implements "link" operation in Kegg REST

        http://www.genome.jp/linkdb/
        """
        url = "http://rest.genome.jp/link/{0}/{1}/{2}".format(
            target,
            source,
            '' if options is None else options
        )

        LOG.debug(url)

        data = url_read(url, agent=self.contact)

        mappings = {}
        for line in data.splitlines():
            source_id, target_id, _ = line.rstrip().split('\t')
            source_id = source_id.split(':')[1]
            target_id = target_id.split(':')[1]
            try:
                mappings[source_id].add(target_id)
            except KeyError:
                mappings[source_id] = set([target_id])

        return mappings

    def find(self, query, database, options=None, strip=True):
        """
        .. versionadded:: 0.3.1

        Kegg Help:

        http://rest.kegg.jp/find/<database>/<query>

        <database> = pathway | module | ko | genome | <org> | compound | glycan |
                     reaction | rclass | enzyme | disease | drug | dgroup | environ |
                     genes | ligand

        <org> = KEGG organism code or T number

        http://rest.kegg.jp/find/<database>/<query>/<option>

        <database> = compound | drug
        <option> = formula | exact_mass | mol_weight

        Examples:
            >>> kc = KeggClientRest()
            >>> kc.find('CH4', 'compound')
            {'C01438': 'Methane; CH4'}
            >>> kc.find('K00844', 'genes', strip=False)
            {'tped:TPE_0072': 'hexokinase; K00844 hexokinase [EC:2.7.1.1]',
            ...
            >>> kc.find('174.05', 'compound', options='exact_mass')
            {'C00493': '174.052823',
             'C04236': '174.052823',
             'C16588': '174.052823',
             'C17696': '174.052823',
             'C18307': '174.052823',
             'C18312': '174.052823',
             'C21281': '174.052823'}
        """

        url = 'http://rest.kegg.jp/find/{}/{}/{}'.format(
            database,
            query,
            '' if options is None else options
        )

        LOG.debug(url)

        data = url_read(url, agent=self.contact)

        mappings = {}
        for line in data.splitlines():
            target_id, description = line.rstrip().split('\t')

            if strip:
                target_id = target_id.split(':')[1]

            mappings[target_id] = description

        return mappings

    def conv(self, target_db, source_db, strip=True):
        """
        .. versionadded:: 0.3.1

        Kegg Help:

        http://rest.kegg.jp/conv/<target_db>/<source_db>

        (<target_db> <source_db>) = (<kegg_db> <outside_db>) | (<outside_db> <kegg_db>)

        For gene identifiers:
        <kegg_db> = <org>
        <org> = KEGG organism code or T number
        <outside_db> = ncbi-proteinid | ncbi-geneid | uniprot

        For chemical substance identifiers:
        <kegg_db> = drug | compound | glycan
        <outside_db> = pubchem | chebi
        http://rest.kegg.jp/conv/<target_db>/<dbentries>

        For gene identifiers:
        <dbentries> = database entries involving the following <database>
        <database> = <org> | genes | ncbi-proteinid | ncbi-geneid | uniprot
        <org> = KEGG organism code or T number

        For chemical substance identifiers:
        <dbentries> = database entries involving the following <database>
        <database> = drug | compound | glycan | pubchem | chebi

        Examples:
            >>> kc = KeggClientRest()
            >>> kc.conv('ncbi-geneid', 'eco')
            {'eco:b0217': {'ncbi-geneid:949009'},
             'eco:b0216': {'ncbi-geneid:947541'},
             'eco:b0215': {'ncbi-geneid:946441'},
             'eco:b0214': {'ncbi-geneid:946955'},
             'eco:b0213': {'ncbi-geneid:944903'},
            ...
            >>> kc.conv('ncbi-proteinid', 'hsa:10458+ece:Z5100')
            {'10458': {'NP_059345'}, 'Z5100': {'AAG58814'}}
        """

        url = 'http://rest.kegg.jp/conv/{}/{}/'.format(
            target_db,
            source_db,
        )

        LOG.debug(url)

        data = url_read(url, agent=self.contact)

        mappings = {}
        for line in data.splitlines():
            source_id, target_id = line.rstrip().split('\t')

            if strip:
                target_id = target_id.split(':')[1]
                source_id = source_id.split(':')[1]

            try:
                mappings[source_id].add(target_id)
            except KeyError:
                mappings[source_id] = set([target_id])

        return mappings

    # names #

    def get_ids_names(self, target='ko', strip=True):
        """
        .. versionadded:: 0.1.13

        .. versionchanged:: 0.3.1
            the call is now cached

        Returns a dictionary with the names/description of all the id of a
        specific target, (ko, path, cpd, etc.)

        If strip=True the id will stripped of the module abbreviation (e.g.
        md:M00002->M00002)
        """

        if strip:
            try:
                return self.cache['get_ids_names'][target].copy()
            except KeyError:
                LOG.debug('No cached values for "%s"', target)

        id_names = {}

        for line in self.list_ids(target).splitlines():
            kegg_id, name = line.strip().split('\t')
            if strip:
                kegg_id = kegg_id.split(':')[1]
            id_names[kegg_id] = name

        if strip:
            self.cache['get_ids_names'][target] = id_names.copy()

        return id_names

    def get_ortholog_pathways(self):
        """
        Gets ortholog pathways, replace 'map' with 'ko' in the id
        """
        data = self.get_ids_names('pathway')
        pathways = {}
        for kegg_id, name in viewitems(data):

            kegg_id = kegg_id.replace('map', 'ko')
            pathways[kegg_id] = name

        return pathways

    # end names #

    def get_reaction_equations(self, ids, max_len=10):
        "Get the equation for the reactions"
        if isinstance(ids, str):
            ids = [ids]
        data = {}
        for idx in range(0, len(ids), max_len):
            if len(ids) > max_len:
                LOG.info(
                    "Downloading reactions equations - range %d-%d",
                    idx + 1, idx + max_len)

            url = "{0}get/{1}".format(
                self.api_url, '+'.join(ids[idx:idx+max_len])
            )

            t_data = url_read(url)
            for entry in t_data.split('///'):
                for line in entry.split('\n'):
                    if line.startswith('EQUATION'):
                        cp_in, cp_out = line.split('<=>')
                        # print cp_in, cp_out
                        cp_in = self.rn_eq_re.findall(cp_in)
                        cp_out = self.rn_eq_re.findall(cp_out)
                        # print cp_in, cp_out
                    elif line.startswith('ENTRY'):
                        # try:
                        name = self.rn_name_re.search(line).group(0)
                        # except:
                        #     print line
                data[name] = {'in': cp_in, 'out': cp_out}

        return data

    def get_pathway_links(self, pathway):
        """
        Returns a dictionary with the mappings KO->compounds for a specific
        Pathway or module
        """
        compounds = self.link_ids('cpd', pathway)
        compounds = set(compounds[pathway])
        kos = self.link_ids('ko', pathway)
        kos = set(kos[pathway])
        cp_rn = self.link_ids('ec', list(compounds))
        ko_rn = self.link_ids('ec', list(kos))
        cp_rn_rev = dict_utils.reverse_mapping(cp_rn)
        ko_cp = dict_utils.combine_dict(ko_rn, cp_rn_rev)

        return ko_cp


BLACK_LIST = [
    'ko05164',  # Influenza A
    'ko05166',  # HTLV-I infection
    'ko05161',  # Hepatitis B
    'ko05160',  # Hepatitis C
    'ko05162',  # Measles
    'ko05169',  # Epstein-Barr virus infection
    'ko05168',  # Herpes simplex infection
    'ko04520',  # Adherens junction
    'ko05014',  # Amyotrophic lateral sclerosis (ALS)
    'ko05016',  # Huntington's disease
    'ko05310',  # Asthma
    'ko00908',  # Zeatin biosynthesis
    'ko04930',  # Type II diabetes mellitus
    'ko04622',  # RIG-I-like receptor signaling pathway
    'ko04626',  # Plant-pathogen interaction
    'ko05222',  # Small cell lung cancer
    'ko05223',  # Non-small cell lung cancer
    'ko05220',  # Chronic myeloid leukemia
    'ko05221',  # Acute myeloid leukemia
    'ko04612',  # Antigen processing and presentation
    'ko04621',  # NOD-like receptor signaling pathway
    'ko04620',  # Toll-like receptor signaling pathway
    'ko04623',  # Cytosolic DNA-sensing pathway
    'ko04622',  # RIG-I-like receptor signaling pathway
    'ko00351',  # DDT degradation
    'ko04270',  # Vascular smooth muscle contraction
    'ko04115',  # p53 signaling pathway
    'ko04114',  # Oocyte meiosis
    'ko05120',  # Epithelial cell signaling in Helicobacter pylori infection
    'ko00590',  # Arachidonic acid metabolism
    'ko04370',  # VEGF signaling pathway
    'ko04976',  # Bile secretion
    'ko04971',  # Gastric acid secretion
    'ko04970',  # Salivary secretion
    'ko04973',  # Carbohydrate digestion and absorption
    'ko04972',  # Pancreatic secretion
    'ko04975',  # Fat digestion and absorption
    'ko04974',  # Protein digestion and absorption
    'ko04977',  # Vitamin digestion and absorption
    'ko04711',  # Circadian rhythm - fly
    'ko04710',  # Circadian rhythm
    'ko04713',  # Circadian entrainment
    'ko04712',  # Circadian rhythm - plant
    'ko04650',  # Natural killer cell mediated cytotoxicity
    'ko04391',  # Hippo signaling pathway - fly
    'ko04151',  # PI3K-Akt signaling pathway
    'ko04150',  # mTOR signaling pathway
    'ko04510',  # Focal adhesion
    'ko04012',  # ErbB signaling pathway
    'ko04013',  # MAPK signaling pathway - fly
    'ko04010',  # MAPK signaling pathway
    'ko03320',  # PPAR signaling pathway
    'ko04530',  # Tight junction
    'ko00981',  # Insect hormone biosynthesis
    'ko00982',  # Drug metabolism - cytochrome P450
    'ko00983',  # Drug metabolism - other enzymes
    'ko05150',  # Staphylococcus aureus infection
    'ko05152',  # Tuberculosis
    'ko04614',  # Renin-angiotensin system
    'ko04610',  # Complement and coagulation cascades
    'ko04340',  # Hedgehog signaling pathway
    'ko00984',  # Steroid degradation
    'ko05010',  # Alzheimer's disease
    'ko05012',  # Parkinson's disease
    'ko04260',  # Cardiac muscle contraction
    'ko04540',  # Gap junction
    'ko05110',  # Vibrio cholerae infection
    'ko05111',  # Vibrio cholerae pathogenic cycle
    'ko04640',  # Hematopoietic cell lineage
    'ko04310',  # Wnt signaling pathway
    'ko04728',  # Dopaminergic synapse
    'ko04724',  # Glutamatergic synapse
    'ko04725',  # Cholinergic synapse
    'ko04726',  # Serotonergic synapse
    'ko04727',  # GABAergic synapse
    'ko04720',  # Long-term potentiation
    'ko04721',  # Synaptic vesicle cycle
    'ko04722',  # Neurotrophin signaling pathway
    'ko04723',  # Retrograde endocannabinoid signaling
    'ko05340',  # Primary immunodeficiency
    'ko05414',  # Dilated cardiomyopathy
    'ko05416',  # Viral myocarditis
    'ko05410',  # Hypertrophic cardiomyopathy (HCM)
    'ko05412',  # Arrhythmogenic right ventricular cardiomyopathy (ARVC)
    'ko04130',  # SNARE interactions in vesicular transport
    'ko05145',  # Toxoplasmosis
    'ko05144',  # Malaria
    'ko04350',  # TGF-beta signaling pathway
    'ko04210',  # Apoptosis
    'ko05200',  # Pathways in cancer
    'ko05202',  # Transcriptional misregulation in cancer
    'ko05203',  # Viral carcinogenesis
    'ko05020',  # Prion diseases
    'ko05143',  # African trypanosomiasis
    'ko05142',  # Chagas disease (American trypanosomiasis)
    'ko05140',  # Leishmaniasis
    'ko00072',  # Synthesis and degradation of ketone bodies
    'ko00073',  # Cutin, suberine and wax biosynthesis
    'ko05330',  # Allograft rejection
    'ko04912',  # GnRH signaling pathway
    'ko05332',  # Graft-versus-host disease
    'ko04910',  # Insulin signaling pathway
    'ko04916',  # Melanogenesis
    'ko04914',  # Progesterone-mediated oocyte maturation
    'ko04672',  # Intestinal immune network for IgA production
    'ko04670',  # Leukocyte transendothelial migration
    'ko04660',  # T cell receptor signaling pathway
    'ko00603',  # Glycosphingolipid biosynthesis - globo series
    'ko00601',  # Glycosphingolipid biosynthesis - lacto and neolacto series
    'ko00600',  # Sphingolipid metabolism
    'ko00604',  # Glycosphingolipid biosynthesis - ganglio series
    'ko04070',  # Phosphatidylinositol signaling system
    'ko04075',  # Plant hormone signal transduction
    'ko05100',  # Bacterial invasion of epithelial cells
    'ko05204',  # Chemical carcinogenesis
    'ko04920',  # Adipocytokine signaling pathway
    'ko04080',  # Neuroactive ligand-receptor interaction
    'ko04320',  # Dorso-ventral axis formation
    'ko04730',  # Long-term depression
    'ko04950',  # Maturity onset diabetes of the young
    'ko04630',  # Jak-STAT signaling pathway
    'ko04742',  # Taste transduction
    'ko04740',  # Olfactory transduction
    'ko04744',  # Phototransduction
    'ko04745',  # Phototransduction - fly
    'ko04512',  # ECM-receptor interaction
    'ko03460',  # Fanconi anemia pathway
    'ko04360',  # Axon guidance
    'ko05219',  # Bladder cancer
    'ko05218',  # Melanoma
    'ko05217',  # Basal cell carcinoma
    'ko05216',  # Thyroid cancer
    'ko05215',  # Prostate cancer
    'ko05214',  # Glioma
    'ko05213',  # Endometrial cancer
    'ko05212',  # Pancreatic cancer
    'ko05211',  # Renal cell carcinoma
    'ko05210',  # Colorectal cancer
    'ko05034',  # Alcoholism
    'ko05033',  # Nicotine addiction
    'ko05032',  # Morphine addiction
    'ko05031',  # Amphetamine addiction
    'ko05030',  # Cocaine addiction
    'ko05134',  # Legionellosis
    'ko05132',  # Salmonella infection
    'ko05133',  # Pertussis
    'ko05130',  # Pathogenic Escherichia coli infection
    'ko05131',  # Shigellosis

    # Too big
    'ko01100',  # Metabolic pathways
    'ko01110',  # Biosynthesis of secondary metabolites
    'ko01120',  # Microbial metabolism in diverse environments

    'ko04662',  # B cell receptor signaling pathway
    'ko04966',  # Collecting duct acid secretion
    'ko05322',  # Systemic lupus erythematosus
    'ko04964',  # Proximal tubule bicarbonate reclamation
    'ko05320',  # Autoimmune thyroid disease
    'ko04962',  # Vasopressin-regulated water reabsorption
    'ko04960',  # Aldosterone-regulated sodium reabsorption
    'ko04961',  # Endocrine and other factor-regulated calcium reabsorption
    'ko04380',  # Osteoclast differentiation
    'ko04664',  # Fc epsilon RI signaling pathway
    'ko04666',  # Fc gamma R-mediated phagocytosis
    'ko04062',  # Chemokine signaling pathway
    'ko04060',  # Cytokine-cytokine receptor interaction
    'ko04066',  # HIF-1 signaling pathway
    'ko04064',  # NF-kappa B signaling pathway
    'ko05323',  # Rheumatoid arthritis
    'ko00140',  # Steroid hormone biosynthesis
]


class KeggModule(object):
    """
    .. versionadded:: 0.1.13

    Used to extract information from a pathway module entry in Kegg

    The entry, as a string, can be either passed at instance creation or with
    :meth:`KeggModule.parse_entry`

    """
    entry = ''
    name = ''
    classes = None
    compounds = None

    _orthologs = None
    _reactions = None
    reactions = None

    def __init__(self, entry=None, old=False):
        """
        .. versionchanged:: 0.3.0
            added *old* parameter, to use the old parser
        """
        if entry is None:
            return
        if old:
            self.parse_entry(entry)
        else:
            self.parse_entry2(entry)

    def parse_entry(self, entry):
        """
        Parses a Kegg module entry and change the instance values. By default
        the reactions IDs are substituted with the KO IDs
        """
        entryd = {}
        curr_field = ''
        for line in entry.splitlines():
            if line.startswith(' '):
                entryd[curr_field].append(line.strip())
            elif line.startswith('///'):
                continue
            else:
                curr_field = line.split(' ')[0]
                entryd[curr_field] = []
                entryd[curr_field].append(line.replace(curr_field, '').strip())

        self.entry = re.search(r"(M\d{5})\s+.+", entryd['ENTRY'][0]).group(1)
        self.name = entryd['NAME'][0]

        self.classes = entryd['CLASS'][0].split('; ')
        self.compounds = [
            re.search(r"(C\d{5})\s+.+", line).group(1)
            for line in entryd['COMPOUND']
        ]

        self.reactions = [
            self.parse_reaction(reaction, ko_ids)
            for ko_ids, reaction in zip(
                entryd['DEFINITION'][0].split(' '),
                entryd['REACTION']
            )
        ]
        self._orthologs = entryd['DEFINITION'][0].split(' ')
        self._reactions = entryd['REACTION']

    def parse_entry2(self, entry):
        """
        .. versionadded:: 0.3.0

        Parses a Kegg module entry and change the instance values. By default
        the reactions IDs are NOT substituted with the KO IDs.
        """
        entryd = {}
        curr_field = ''
        for line in entry.splitlines():
            if line.startswith(' '):
                entryd[curr_field].append(line.strip())
            elif line.startswith('///'):
                continue
            else:
                curr_field = line.split(' ')[0]
                entryd[curr_field] = []
                entryd[curr_field].append(line.replace(curr_field, '').strip())

        self.entry = re.search(r"(M\d{5})\s+.+", entryd['ENTRY'][0]).group(1)
        self.name = entryd['NAME'][0]

        self.classes = entryd['CLASS'][0].split('; ')
        self.compounds = [
            re.search(r"(C\d{5})\s+.+", line).group(1)
            for line in entryd['COMPOUND']
        ]

        self.reactions = []

        for reaction in entryd['REACTION']:
            reaction = self.parse_reaction(reaction, ko_ids=None)
            self.reactions.append(reaction)

        self._orthologs = entryd['DEFINITION'][0].split(' ')
        self._reactions = entryd['REACTION']

    @staticmethod
    def parse_reaction(line, ko_ids=None):
        """
        .. versionchanged:: 0.3.0
            cleaned the parsing

        parses the lines with the reactions and substitute reaction IDs with
        the corresponding KO IDs if provided
        """

        # line = 'R00294,R02492,R09446,R09808,R09809  C00533 -> C00887'
        # ko_ids = '(K00370+K00371+K00374+K00373,K02567+K02568)'
        # some reaction lines have only one space
        rn_ids, reaction = line.replace('  ', ' ').split(' ', 1)
        rn_ids = tuple(x.strip() for x in rn_ids.replace('+', ',').split(','))
        comp1, comp2 = reaction.split(' -> ')
        comp1 = comp1.replace('(spontaneous)', '')
        comp2 = comp2.replace('(spontaneous)', '')
        comp1 = tuple(x.strip() for x in comp1.replace('+', ',').split(','))
        comp2 = tuple(x.strip() for x in comp2.replace('+', ',').split(','))
        if ko_ids is not None:
            rn_ids = ko_ids.replace('+', ',').replace('-', ',').replace('(', '').replace(')', '').split(',')
        return rn_ids, (comp1, comp2)

    @property
    def first_cp(self):
        "Returns the first compound in the module"
        return self.reactions[0][1][0][0]

    @property
    def last_cp(self):
        "Returns the first compound in the module"
        return self.reactions[-1][-1][-1][0]

    def to_edges(self, id_only=None):
        """
        .. versionchanged:: 0.3.0
            added id_only and changed to reflect changes in :attr:`reactions`

        Returns the reactions as edges that can be supplied to make graph.

        Arguments:
            id_only (None, iterable): if None the returned edges are for the
                whole module, if an iterable (converted to a :class:`set`),
                only edges for those reactions are returned

        Yield:
            tuple: the elements are the compounds and reactions in the module
        """

        if id_only is not None:
            id_only = set(id_only)

        for rn_ids, (comp1s, comp2s) in self.reactions:
            for rn_id in rn_ids:
                if (id_only is not None) and (rn_id not in id_only):
                    continue
                for comp1 in comp1s:
                    yield (comp1, rn_id)
                for comp2 in comp2s:
                    yield (rn_id, comp2)

    def find_submodules(self):
        """
        .. versionadded:: 0.3.0

        Returns the possible submodules, as a list of tuples where the elements
        are the first and last compounds in a submodule
        """
        sub_modules = []
        sub_module = None
        for rn_ids, (left_cpds, right_cpds) in self.reactions:
            if sub_module is None:
                sub_module = [left_cpds, right_cpds]
                continue

            if set(sub_module[1]) & set(left_cpds):
                sub_module[1] = right_cpds
            else:
                sub_modules.append((sub_module[0][0], sub_module[-1][-1]))
                sub_module = [left_cpds, right_cpds]
        else:
            sub_modules.append((sub_module[0][0], sub_module[-1][-1]))
        return sub_modules


def parse_reaction(line, prefix=('C', 'G')):
    """
    .. versionadded:: 0.3.1

    Parses a reaction equation from Kegg, returning the left and right
    components. Needs testing

    Arguments:
        line (str): reaction string

    Returns:
        tuple: left and right components as `sets`

    Raises:
        ValueError: if the

    """
    line = line.replace('EQUATION', '').strip()
    if '<=>' in line:
        line = line.replace(' ', '').split('<=>')
        left = set(x if x.startswith('C') else x[1:] for x in line[0].split('+'))
        right = set(x if x.startswith('C') else x[1:] for x in line[1].split('+'))
        return left, right
    elif '=>' in line:
        raise ValueError('>>>')
    elif '<=' in line:
        raise ValueError('<<<')
    else:
        raise ValueError('???')
